"""Utils for ASR functions."""
import numpy as np
from numpy import linalg
from scipy import signal
from scipy.linalg import toeplitz
from scipy.spatial.distance import cdist, euclidean
from scipy.special import gamma, gammaincinv

SHAPE_RANGE = np.linspace(1.7, 3.5, 13)


def fit_eeg_distribution(X, min_clean_fraction=0.25, max_dropout_fraction=0.1,
                         fit_quantiles=[0.022, 0.6], step_sizes=[0.0220, 0.6000],
                         shape_range=SHAPE_RANGE):
    """Estimate the mean and SD of clean EEG from contaminated data.

    This function estimates the mean and standard deviation of clean EEG from a
    sample of amplitude values (that have preferably been computed over short
    windows) that may include a large fraction of contaminated samples. The
    clean EEG is assumed to represent a generalized Gaussian component in a
    mixture with near-arbitrary artifact components. By default, at least 25%
    (``min_clean_fraction``) of the data must be clean EEG, and the rest can be
    contaminated. No more than 10% (``max_dropout_fraction``) of the data is
    allowed to come from contaminations that cause lower-than-EEG amplitudes
    (e.g., sensor unplugged). There are no restrictions on artifacts causing
    larger-than-EEG amplitudes, i.e., virtually anything is handled (with the
    exception of a very unlikely type of distribution that combines with the
    clean EEG samples into a larger symmetric generalized Gaussian peak and
    thereby "fools" the estimator). The default parameters should work for a
    wide range of applications but may be adapted to accommodate special
    circumstances.

    The method works by fitting a truncated generalized Gaussian whose
    parameters are constrained by ``min_clean_fraction``,
    ``max_dropout_fraction``, ``fit_quantiles``, and ``shape_range``. The fit
    is performed by a grid search that always finds a close-to-optimal solution
    if the above assumptions are fulfilled.

    Parameters
    ----------
    X : array, shape=(n_channels, n_samples)
        EEG data, possibly containing artifacts.
    min_clean_fraction : float
        Minimum fraction that needs to be clean. This is the minimum fraction
        of time windows that need to contain essentially uncontaminated EEG
        (default=0.1).
    max_dropout_fraction : float
        Maximum fraction that can have dropouts. This is the maximum fraction
        of time windows that may have arbitrarily low amplitude (e.g., due to
        the sensors being unplugged) (default=0.25).
    fit_quantiles : 2-tuple
        Quantile range [lower,upper] of the truncated generalized Gaussian
        distribution that shall be fit to the EEG contents (default=[0.022
        0.6]).
    step_sizes : 2-tuple
        Step size of the grid search; the first value is the stepping of the
        lower bound (which essentially steps over any dropout samples), and the
        second value is the stepping over possible scales (i.e., clean-data
        quantiles) (default=[0.01, 0.01]).
    shape_range : array
        Range that the clean EEG distribution's shape parameter beta may take.

    Returns
    -------
    mu : array
        Estimated mean of the clean EEG distribution.
    sig : array
        Estimated standard deviation of the clean EEG distribution.
    alpha : float
        Estimated scale parameter of the generalized Gaussian clean EEG
        distribution.
    beta : float
        Estimated shape parameter of the generalized Gaussian clean EEG
        distribution.

    """
    # sort data so we can access quantiles directly
    X = np.sort(X)
    n = len(X)

    # compute z bounds for the truncated standard generalized Gaussian pdf and
    # pdf rescaler
    quants = np.array(fit_quantiles)
    zbounds = []
    rescale = []
    for b in range(len(shape_range)):
        gam = gammaincinv(
            1 / shape_range[b], np.sign(quants - 1 / 2) * (2 * quants - 1))
        zbounds.append(np.sign(quants - 1 / 2) * gam ** (1 / shape_range[b]))
        rescale.append(shape_range[b] / (2 * gamma(1 / shape_range[b])))

    # determine the quantile-dependent limits for the grid search
    # we can generally skip the tail below the lower quantile
    lower_min = np.min(quants)
    # maximum width is the fit interval if all data is clean
    max_width = np.diff(quants)
    # minimum width of the fit interval, as fraction of data
    min_width = min_clean_fraction * max_width

    # Build quantile interval matrix
    cols = np.arange(lower_min,
                     lower_min + max_dropout_fraction + step_sizes[0] * 1e-9,
                     step_sizes[0])
    cols = np.round(n * cols).astype(int)
    rows = np.arange(0, int(np.round(n * max_width)))
    newX = np.zeros((len(rows), len(cols)))
    for i, c in enumerate(range(len(rows))):
        newX[i] = X[c + cols]

    # subtract baseline value for each interval
    X1 = newX[0, :]
    newX = newX - X1
    opt_val = np.inf

    opt_val = np.inf
    opt_lu = np.inf
    opt_bounds = np.inf
    opt_beta = np.inf
    gridsearch = np.round(n * np.arange(max_width, min_width, -step_sizes[1]))
    for m in gridsearch.astype(int):
        mcurr = m - 1
        nbins = int(np.round(3 * np.log2(1 + m / 2)))
        cols = nbins / newX[mcurr]
        H = newX[:m] * cols

        hist_all = []
        for ih in range(len(cols)):
            histcurr = np.histogram(H[:, ih], bins=np.arange(0, nbins + 1))
            hist_all.append(histcurr[0])
        hist_all = np.array(hist_all, dtype=int).T
        hist_all = np.vstack((hist_all, np.zeros(len(cols), dtype=int)))
        logq = np.log(hist_all + 0.01)

        # for each shape value...
        for k, b in enumerate(shape_range):
            bounds = zbounds[k]
            x = bounds[0] + np.arange(0.5, nbins + 0.5) / nbins * np.diff(bounds)  # noqa:E501
            p = np.exp(-np.abs(x) ** b) * rescale[k]
            p = p / np.sum(p)

            # calc KL divergences
            kl = np.sum(p * (np.log(p) - logq[:-1, :].T), axis=1) + np.log(m)

            # update optimal parameters
            min_val = np.min(kl)
            idx = np.argmin(kl)
            if min_val < opt_val:
                opt_val = min_val
                opt_beta = shape_range[k]
                opt_bounds = bounds
                opt_lu = [X1[idx], X1[idx] + newX[m - 1, idx]]

    # recover distribution parameters at optimum
    alpha = (opt_lu[1] - opt_lu[0]) / np.diff(opt_bounds)
    mu = opt_lu[0] - opt_bounds[0] * alpha
    beta = opt_beta

    # calculate the distribution's standard deviation from alpha and beta
    sig = np.sqrt((alpha ** 2) * gamma(3 / beta) / gamma(1 / beta))

    return mu, sig, alpha, beta


def yulewalk(order, F, M):
    """Recursive filter design using a least-squares method.

    [B,A] = YULEWALK(N,F,M) finds the N-th order recursive filter
    coefficients B and A such that the filter:

    B(z)   b(1) + b(2)z^-1 + .... + b(n)z^-(n-1)
    ---- = -------------------------------------
    A(z)    1   + a(1)z^-1 + .... + a(n)z^-(n-1)

    matches the magnitude frequency response given by vectors F and M.

    The YULEWALK function performs a least squares fit in the time domain. The
    denominator coefficients {a(1),...,a(NA)} are computed by the so called
    "modified Yule Walker" equations, using NR correlation coefficients
    computed by inverse Fourier transformation of the specified frequency
    response H.

    The numerator is computed by a four step procedure. First, a numerator
    polynomial corresponding to an additive decomposition of the power
    frequency response is computed. Next, the complete frequency response
    corresponding to the numerator and denominator polynomials is evaluated.
    Then a spectral factorization technique is used to obtain the impulse
    response of the filter. Finally, the numerator polynomial is obtained by a
    least squares fit to this impulse response. For a more detailed explanation
    of the algorithm see [1]_.

    Parameters
    ----------
    order : int
        Filter order.
    F : array
        Normalised frequency breakpoints for the filter. The frequencies in F
        must be between 0.0 and 1.0, with 1.0 corresponding to half the sample
        rate. They must be in increasing order and start with 0.0 and end with
        1.0.
    M : array
        Magnitude breakpoints for the filter such that PLOT(F,M) would show a
        plot of the desired frequency response.

    References
    ----------
    .. [1] B. Friedlander and B. Porat, "The Modified Yule-Walker Method of
           ARMA Spectral Estimation," IEEE Transactions on Aerospace Electronic
           Systems, Vol. AES-20, No. 2, pp. 158-173, March 1984.

    Examples
    --------
    Design an 8th-order lowpass filter and overplot the desired
    frequency response with the actual frequency response:

    >>> f = [0, .6, .6, 1]         # Frequency breakpoints
    >>> m = [1, 1, 0, 0]           # Magnitude breakpoints
    >>> [b, a] = yulewalk(8, f, m) # Filter design using a least-squares method

    """
    F = np.asarray(F)
    M = np.asarray(M)
    npt = 512
    lap = np.fix(npt / 25).astype(int)
    mf = F.size
    npt = npt + 1  # For [dc 1 2 ... nyquist].
    Ht = np.array(np.zeros((1, npt)))
    nint = mf - 1
    df = np.diff(F)

    nb = 0
    Ht[0][0] = M[0]
    for i in range(nint):
        if df[i] == 0:
            nb = nb - int(lap / 2)
            ne = nb + lap
        else:
            ne = int(np.fix(F[i + 1] * npt)) - 1

        j = np.arange(nb, ne + 1)
        if ne == nb:
            inc = 0
        else:
            inc = (j - nb) / (ne - nb)

        Ht[0][nb:ne + 1] = np.array(inc * M[i + 1] + (1 - inc) * M[i])
        nb = ne + 1

    Ht = np.concatenate((Ht, Ht[0][-2:0:-1]), axis=None)
    n = Ht.size
    n2 = np.fix((n + 1) / 2)
    nb = order
    nr = 4 * order
    nt = np.arange(0, nr)

    # compute correlation function of magnitude squared response
    R = np.real(np.fft.ifft(Ht * Ht))
    R = R[0:nr] * (0.54 + 0.46 * np.cos(np.pi * nt / (nr - 1)))   # pick NR correlations  # noqa

    # Form window to be used in extracting the right "wing" of two-sided
    # covariance sequence
    Rwindow = np.concatenate(
        (1 / 2, np.ones((1, int(n2 - 1))), np.zeros((1, int(n - n2)))),
        axis=None)
    A = polystab(denf(R, order))  # compute denominator

    # compute additive decomposition
    Qh = numf(np.concatenate((R[0] / 2, R[1:nr]), axis=None), A, order)

    # compute impulse response
    _, Ss = 2 * np.real(signal.freqz(Qh, A, worN=n, whole=True))

    hh = np.fft.ifft(
        np.exp(np.fft.fft(Rwindow * np.fft.ifft(np.log(Ss, dtype=complex))))
    )
    B = np.real(numf(hh[0:nr], A, nb))

    return B, A


def yulewalk_filter(X, sfreq, zi=None, ab=None, axis=-1):
    """Yulewalk filter.

    Parameters
    ----------
    X : array, shape = (n_channels, n_samples)
        Data to filter.
    sfreq : float
        Sampling frequency.
    zi : array, shape=(n_channels, filter_order)
        Initial conditions.
    a, b : 2-tuple | None
        Coefficients of an IIR filter that is used to shape the spectrum of the
        signal when calculating artifact statistics. The output signal does not
        go through this filter. This is an optional way to tune the sensitivity
        of the algorithm to each frequency component of the signal. The default
        filter is less sensitive at alpha and beta frequencies and more
        sensitive at delta (blinks) and gamma (muscle) frequencies.
    axis : int
        Axis to filter on (default=-1, corresponding to samples).

    Returns
    -------
    out : array
        Filtered data.
    zf :  array, shape=(n_channels, filter_order)
        Output filter state.

    """
    [C, S] = X.shape
    if ab is None:
        F = np.array([0, 2, 3, 13, 16, 40, np.minimum(
            80.0, (sfreq / 2.0) - 1.0), sfreq / 2.0]) * 2.0 / sfreq
        M = np.array([3, 0.75, 0.33, 0.33, 1, 1, 3, 3])
        B, A = yulewalk(8, F, M)
    else:
        A, B = ab

    # apply the signal shaping filter and initialize the IIR filter state
    if zi is None:
        zi = signal.lfilter_zi(B, A)
        zi = np.transpose(X[:, 0] * zi[:, None])
        out, zf = signal.lfilter(B, A, X, zi=zi, axis=axis)
    else:
        out, zf = signal.lfilter(B, A, X, zi=zi, axis=axis)

    return out, zf


def geometric_median(X, tol=1e-5, max_iter=500):
    """Geometric median.

    This code is adapted from [2]_ using the Vardi and Zhang algorithm
    described in [1]_.

    Parameters
    ----------
    X : array, shape=(n_observations, n_variables)
        The data.
    tol : float
        Tolerance (default=1.e-5)
    max_iter : int
        Max number of iterations (default=500):

    Returns
    -------
    y1 : array, shape=(n_variables,)
        Geometric median over X.

    References
    ----------
    .. [1] Vardi, Y., & Zhang, C. H. (2000). The multivariate L1-median and
       associated data depth. Proceedings of the National Academy of Sciences,
       97(4), 1423-1426. https://doi.org/10.1073/pnas.97.4.1423
    .. [2] https://stackoverflow.com/questions/30299267/
    """
    y = np.mean(X, 0)  # initial value

    i = 0
    while i < max_iter:
        D = cdist(X, [y])
        nonzeros = (D != 0)[:, 0]

        Dinv = 1. / D[nonzeros]
        Dinvs = np.sum(Dinv)
        W = Dinv / Dinvs
        T = np.sum(W * X[nonzeros], 0)

        num_zeros = len(X) - np.sum(nonzeros)
        if num_zeros == 0:
            y1 = T
        elif num_zeros == len(X):
            return y
        else:
            R = (T - y) * Dinvs
            r = np.linalg.norm(R)
            rinv = 0 if r == 0 else num_zeros / r
            y1 = max(0, 1 - rinv) * T + min(1, rinv) * y

        if euclidean(y, y1) < tol:
            return y1

        y = y1
        i += 1
    else:
        print(f"Geometric median could converge in {i} iterations "
              f"with a tolerance of {tol}")


def polystab(a):
    """Polynomial stabilization.

    POLYSTAB(A), where A is a vector of polynomial coefficients,
    stabilizes the polynomial with respect to the unit circle;
    roots whose magnitudes are greater than one are reflected
    inside the unit circle.

    Examples
    --------
    Convert a linear-phase filter into a minimum-phase filter with the same
    magnitude response.

    >>> h = fir1(25,0.4);               # Window-based FIR filter design
    >>> flag_linphase = islinphase(h)   # Determines if filter is linear phase
    >>> hmin = polystab(h) * norm(h)/norm(polystab(h));
    >>> flag_minphase = isminphase(hmin)# Determines if filter is minimum phase

    """
    v = np.roots(a)
    i = np.where(v != 0)
    vs = 0.5 * (np.sign(np.abs(v[i]) - 1) + 1)
    v[i] = (1 - vs) * v[i] + vs / np.conj(v[i])
    ind = np.where(a != 0)
    b = a[ind[0][0]] * np.poly(v)

    # Return only real coefficients if input was real:
    if not np.sum(np.imag(a)):
        b = np.real(b)

    return b


def numf(h, a, nb):
    """Find numerator B given impulse-response h of B/A and denominator A.

    NB is the numerator order.  This function is used by YULEWALK.
    """
    nh = np.max(h.size)
    xn = np.concatenate((1, np.zeros((1, nh - 1))), axis=None)
    impr = signal.lfilter(np.array([1.0]), a, xn)

    b = linalg.lstsq(
        toeplitz(impr, np.concatenate((1, np.zeros((1, nb))), axis=None)),
        h.T, rcond=None)[0].T

    return b


def denf(R, na):
    """Compute denominator from covariances.

    A = DENF(R,NA) computes order NA denominator A from covariances
    R(0)...R(nr) using the Modified Yule-Walker method. This function is used
    by YULEWALK.

    """
    nr = np.max(np.size(R))
    Rm = toeplitz(R[na:nr - 1], R[na:0:-1])
    Rhs = - R[na + 1:nr]
    A = np.concatenate(
        (1, linalg.lstsq(Rm, Rhs.T, rcond=None)[0].T), axis=None)
    return A



"""Synthetic test data."""
import matplotlib.pyplot as plt
import numpy as np

from meegkit.utils import fold, rms, unfold


def create_line_data(n_samples=100 * 3, n_chans=30, n_trials=100, noise_dim=20,
                     n_bad_chans=1, SNR=.1, fline=1, t0=None, show=False):
    """Create synthetic data.

    Parameters
    ----------
    n_samples : int
        Number of samples (default=100*3).
    n_chans : int
        Number of channels (default=30).
    n_trials : int
        Number of trials (default=100).
    noise_dim : int
        Dimensionality of noise (default=20).
    n_bad_chans : int
        Number of bad channels (default=1).
    t0 : int
        Onset sample of artifact.
    fline : float
        Normalized frequency of artifact (freq/samplerate), (default=1).

    Returns
    -------
    data : ndarray, shape=(n_samples, n_chans, n_trials)
    source : ndarray, shape=(n_samples,)
    """
    rng = np.random.RandomState(2022)

    if t0 is None:
        t0 = n_samples // 3
    t1 = n_samples - 2 * t0  # artifact duration

    # create source signal
    source = np.hstack((
        np.zeros(t0),
        np.sin(2 * np.pi * fline * np.arange(t1)),
        np.zeros(t0)))  # noise -> artifact -> noise
    source = source[:, None]

    # mix source in channels
    s = source * rng.randn(1, n_chans)
    s = s[:, :, np.newaxis]
    s = np.tile(s, (1, 1, n_trials))  # create trials

    # set first `n_bad_chans` to zero
    s[:, :n_bad_chans] = 0.

    # noise
    noise = np.dot(
        unfold(rng.randn(n_samples, noise_dim, n_trials)),
        rng.randn(noise_dim, n_chans))
    noise = fold(noise, n_samples)

    # mix signal and noise
    data = noise / rms(noise.flatten()) + SNR * s / rms(s.flatten())

    if show:
        f, ax = plt.subplots(3)
        ax[0].plot(source.mean(-1), label="source")
        ax[1].plot(noise[:, 1].mean(-1), label="noise (avg over trials)")
        ax[2].plot(data[:, 1].mean(-1), label="mixture (avg over trials)")
        ax[0].legend()
        ax[1].legend()
        ax[2].legend()
        plt.show()

    return data, source



"""Artifact Subspace Reconstruction."""
import logging

import numpy as np
from scipy import linalg, signal
from statsmodels.robust.scale import mad

from .utils import block_covariance, nonlinear_eigenspace
from .utils.asr import fit_eeg_distribution, geometric_median, yulewalk, yulewalk_filter

try:
    import pyriemann
except ImportError:
    pyriemann = None


class ASR:
    """Artifact Subspace Reconstruction.

    Artifact subspace reconstruction (ASR) is an automatic, online,
    component-based artifact removal method for removing transient or
    large-amplitude artifacts in multi-channel EEG recordings [1]_.

    Parameters
    ----------
    sfreq : float
        Sampling rate of the data, in Hz.

    The following are optional parameters (the key parameter of the method is
    the ``cutoff``):

    cutoff: float
        Standard deviation cutoff for rejection. X portions whose variance
        is larger than this threshold relative to the calibration data are
        considered missing data and will be removed. The most aggressive value
        that can be used without losing too much EEG is 2.5. A quite
        conservative value would be 5 (default=5).
    blocksize : int
        Block size for calculating the robust data covariance and thresholds,
        in samples; allows to reduce the memory and time requirements of the
        robust estimators by this factor (down to Channels x Channels x Samples
        x 16 / Blocksize bytes) (default=10).
    win_len : float
        Window length (s) that is used to check the data for artifact content.
        This is ideally as long as the expected time scale of the artifacts but
        not shorter than half a cycle of the high-pass filter that was used
        (default=1).
    win_overlap : float
        Window overlap fraction. The fraction of two successive windows that
        overlaps. Higher overlap ensures that fewer artifact portions are going
        to be missed, but is slower (default=0.66).
    max_dropout_fraction : float
        Maximum fraction of windows that can be subject to signal dropouts
        (e.g., sensor unplugged), used for threshold estimation (default=0.1).
    min_clean_fraction : float
        Minimum fraction of windows that need to be clean, used for threshold
        estimation (default=0.25).
    method : {'riemann', 'euclid'}
        Method to use. If riemann, use the riemannian-modified version of
        ASR [2]_.
    memory : float
        Memory size (s), regulates the number of covariance matrices to store.
    estimator : str in {'scm', 'lwf', 'oas', 'mcd'}
        Covariance estimator (default: 'scm' which computes the sample
        covariance). Use 'lwf' if you need regularization (requires pyriemann).

    Attributes
    ----------
    ``state_`` : dict
        Initial state of the ASR filter.
    ``zi_``: array, shape=(n_channels, filter_order)
        Filter initial conditions.
    ``ab_``: 2-tuple
        Coefficients of an IIR filter that is used to shape the spectrum of the
        signal when calculating artifact statistics. The output signal does not
        go through this filter. This is an optional way to tune the sensitivity
        of the algorithm to each frequency component of the signal. The default
        filter is less sensitive at alpha and beta frequencies and more
        sensitive at delta (blinks) and gamma (muscle) frequencies.
    ``cov_`` : array, shape=(channels, channels)
        Previous covariance matrix.
    ``state_`` : dict
        Previous ASR parameters (as derived by :func:`asr_calibrate`) for
        successive calls to :meth:`transform`. Required fields are:

        - ``M`` : Mixing matrix
        - ``T`` : Threshold matrix
        - ``R`` : Reconstruction matrix (array | None)

    References
    ----------
    .. [1] Kothe, C. A. E., & Jung, T. P. (2016). U.S. Patent Application No.
       14/895,440. https://patents.google.com/patent/US20160113587A1/en
    .. [2] Blum, S., Jacobsen, N. S. J., Bleichner, M. G., & Debener, S.
       (2019). A Riemannian Modification of Artifact Subspace Reconstruction
       for EEG Artifact Handling. Frontiers in Human Neuroscience, 13.
       https://doi.org/10.3389/fnhum.2019.00141

    """

    def __init__(self, sfreq=250, cutoff=5, blocksize=100, win_len=0.5,
                 win_overlap=0.66, max_dropout_fraction=0.1,
                 min_clean_fraction=0.25, name="asrfilter", method="euclid",
                 estimator="scm", **kwargs):

        if pyriemann is None and method == "riemann":
            logging.warning("Need pyriemann to use riemannian ASR flavor.")
            method = "euclid"

        self.cutoff = cutoff
        self.blocksize = blocksize
        self.win_len = win_len
        self.win_overlap = win_overlap
        self.max_dropout_fraction = max_dropout_fraction
        self.min_clean_fraction = min_clean_fraction
        self.max_bad_chans = 0.3
        self.method = method
        self.memory = int(2 * sfreq)  # smoothing window for covariances
        self.sample_weight = np.geomspace(0.05, 1, num=self.memory + 1)
        self.sfreq = sfreq
        self.estimator = estimator

        self.reset()

    def reset(self):
        """Reset filter."""
        # Initialise yulewalk-filter coefficients with sensible defaults
        F = np.array([0, 2, 3, 13, 16, 40,
                      np.minimum(80.0, (self.sfreq / 2.0) - 1.0),
                      self.sfreq / 2.0]) * 2.0 / self.sfreq
        M = np.array([3, 0.75, 0.33, 0.33, 1, 1, 3, 3])
        B, A = yulewalk(8, F, M)
        self.ab_ = (A, B)
        self.cov_ = []
        self.zi_ = None
        self.state_ = {}
        self._counter = []
        self._fitted = False

    def fit(self, X, y=None, **kwargs):
        """Calibration for the Artifact Subspace Reconstruction method.

        The input to this data is a multi-channel time series of calibration
        data. In typical uses the calibration data is clean resting EEG data of
        data if the fraction of artifact content is below the breakdown point
        of the robust statistics used for estimation (50% theoretical, ~30%
        practical). If the data has a proportion of more than 30-50% artifacts
        then bad time windows should be removed beforehand. This data is used
        to estimate the thresholds that are used by the ASR processing function
        to identify and remove artifact components.

        The calibration data must have been recorded for the same cap design
        from which data for cleanup will be recorded, and ideally should be
        from the same session and same subject, but it is possible to reuse the
        calibration data from a previous session and montage to the extent that
        the cap is placed in the same location (where loss in accuracy is more
        or less proportional to the mismatch in cap placement).

        Parameters
        ----------
        X : array, shape=(n_channels, n_samples)
            The calibration data should have been high-pass filtered (for
            example at 0.5Hz or 1Hz using a Butterworth IIR filter), and be
            reasonably clean not less than 30 seconds (this method is typically
            used with 1 minute or more).

        """
        if X.ndim == 3:
            X = X.squeeze()

        # Find artifact-free windows first
        clean, sample_mask = clean_windows(
            X,
            sfreq=self.sfreq,
            win_len=self.win_len,
            win_overlap=self.win_overlap,
            max_bad_chans=self.max_bad_chans,
            min_clean_fraction=self.min_clean_fraction,
            max_dropout_fraction=self.max_dropout_fraction)

        # Perform calibration
        M, T = asr_calibrate(
            clean,
            sfreq=self.sfreq,
            cutoff=self.cutoff,
            blocksize=self.blocksize,
            win_len=self.win_len,
            win_overlap=self.win_overlap,
            max_dropout_fraction=self.max_dropout_fraction,
            min_clean_fraction=self.min_clean_fraction,
            method=self.method,
            estimator=self.estimator)

        self.state_ = dict(M=M, T=T, R=None)
        self._fitted = True

        return clean, sample_mask

    def transform(self, X, y=None, **kwargs):
        """Apply Artifact Subspace Reconstruction.

        Parameters
        ----------
        X : array, shape=([n_trials, ]n_channels, n_samples)
            Raw data.

        Returns
        -------
        out : array, shape=([n_trials, ]n_channels, n_samples)
            Filtered data.

        """
        if X.ndim == 3:
            if X.shape[0] == 1:  # single epoch case
                out = self.transform(X[0])
                return out[None, ...]
            else:
                outs = [self.transform(x) for x in X]
                return np.stack(outs, axis=0)
        else:
            # Yulewalk-filtered data
            X_filt, self.zi_ = yulewalk_filter(
                X, sfreq=self.sfreq, ab=self.ab_, zi=self.zi_)

        if not self._fitted:
            logging.warning("ASR is not fitted ! Returning unfiltered data.")
            return X

        if self.estimator == "scm":
            cov = 1 / X.shape[-1] * X_filt @ X_filt.T
        else:
            cov = pyriemann.estimation.covariances(X_filt[None, ...],
                                                   self.estimator)[0]

        self._counter.append(X_filt.shape[-1])
        self.cov_.append(cov)

        # Regulate the number of covariance matrices that are stored
        while np.sum(self._counter) > self.memory:
            if len(self.cov_) > 1:
                self.cov_.pop(0)
                self._counter.pop(0)
            else:
                self._counter = [self.memory, ]
                break

        # Exponential covariance weights – the most recent covariance has a
        # weight of 1, while the oldest one in memory has a weight of 5%
        weights = [1, ]
        for c in np.cumsum(self._counter[1:]):
            weights = [self.sample_weight[-c]] + weights

        # Clean data, using covariances weighted by sample_weight
        out, self.state_ = asr_process(X, X_filt, self.state_,
                                       cov=np.stack(self.cov_),
                                       method=self.method,
                                       sample_weight=weights)

        return out


def clean_windows(X, sfreq, max_bad_chans=0.2, zthresholds=[-3.5, 5],
                  win_len=.5, win_overlap=0.66, min_clean_fraction=0.25,
                  max_dropout_fraction=0.1, show=False):
    """Remove periods with abnormally high-power content from continuous data.

    This function cuts segments from the data which contain high-power
    artifacts. Specifically, only windows are retained which have less than a
    certain fraction of "bad" channels, where a channel is bad in a window if
    its power is above or below a given upper/lower threshold (in standard
    deviations from a robust estimate of the EEG power distribution in the
    channel).

    Parameters
    ----------
    X : array, shape=(n_channels, n_samples)
        Continuous data set, assumed to be appropriately high-passed (e.g. >
        1Hz or 0.5Hz - 2.0Hz transition band)
    max_bad_chans : float
        The maximum number or fraction of bad channels that a retained window
        may still contain (more than this and it is removed). Reasonable range
        is 0.05 (very clean output) to 0.3 (very lax cleaning of only coarse
        artifacts) (default=0.2).
    zthresholds : 2-tuple
        The minimum and maximum standard deviations within which the power of
        a channel must lie (relative to a robust estimate of the clean EEG
        power distribution in the channel) for it to be considered "not bad".
        (default=[-3.5, 5]).

    The following are detail parameters that usually do not have to be tuned.
    If you can't get the function to do what you want, you might consider
    adapting these to your data.

    win_len : float
        Window length that is used to check the data for artifact content.
        This is ideally as long as the expected time scale of the artifacts
        but not shorter than half a cycle of the high-pass filter that was
        used. Default: 1.
    win_overlap : float
        Window overlap fraction. The fraction of two successive windows that
        overlaps. Higher overlap ensures that fewer artifact portions are
        going to be missed, but is slower (default=0.66).
    min_clean_fraction : float
        Minimum fraction that needs to be clean. This is the minimum fraction
        of time windows that need to contain essentially uncontaminated EEG.
        (default=0.25)
    max_dropout_fraction : float
        Maximum fraction that can have dropouts. This is the maximum fraction
        of time windows that may have arbitrarily low amplitude (e.g., due to
        the sensors being unplugged) (default=0.1).

    Returns
    -------
    clean : array, shape=(n_channels, n_samples)
        Dataset with bad time periods removed.
    sample_mask : boolean array, shape=(1, n_samples)
        Mask of retained samples (logical array).

    """
    assert 0 < max_bad_chans < 1, "max_bad_chans must be a fraction !"

    # set internal variables
    truncate_quant = [0.0220, 0.6000]
    step_sizes = [0.01, 0.01]
    shape_range = np.arange(1.7, 3.5, 0.15)
    max_bad_chans = np.round(X.shape[0] * max_bad_chans)

    # set data indices
    [nc, ns] = X.shape
    N = int(win_len * sfreq)
    offsets = np.round(np.arange(0, ns - N, (N * (1 - win_overlap))))
    offsets = offsets.astype(int)
    logging.debug("[ASR] Determining channel-wise rejection thresholds")

    wz = np.zeros((nc, len(offsets)))
    for ichan in range(nc):

        # compute root mean squared amplitude
        x = X[ichan, :] ** 2
        Y = np.array([np.sqrt(np.sum(x[o:o + N]) / N) for o in offsets])

        # fit a distribution to the clean EEG part
        mu, sig, alpha, beta = fit_eeg_distribution(
            Y, min_clean_fraction, max_dropout_fraction, truncate_quant,
            step_sizes, shape_range)
        # calculate z scores
        wz[ichan] = (Y - mu) / sig

    # sort z scores into quantiles
    wz[np.isnan(wz)] = np.inf  # Nan to inf
    swz = np.sort(wz, axis=0)

    # determine which windows to remove
    if np.max(zthresholds) > 0:
        mask1 = swz[-(int(max_bad_chans) + 1), :] > np.max(zthresholds)
    if np.min(zthresholds) < 0:
        mask2 = (swz[1 + int(max_bad_chans - 1), :] < np.min(zthresholds))

    bad_by_mad = mad(wz, c=1, axis=0) < .1
    bad_by_std = np.std(wz, axis=0) < .1
    mask3 = np.logical_or(bad_by_mad, bad_by_std)

    # combine the three masks
    remove_mask = np.logical_or.reduce((mask1, mask2, mask3))
    removed_wins = np.where(remove_mask)[0]

    # reconstruct the samples to remove
    sample_maskidx = []
    for i, win in enumerate(removed_wins):
        if i == 0:
            sample_maskidx = np.arange(offsets[win], offsets[win] + N)
        else:
            sample_maskidx = np.r_[(sample_maskidx,
                                    np.arange(offsets[win], offsets[win] + N))]

    # delete the bad chunks from the data
    sample_mask2remove = np.unique(sample_maskidx)
    if sample_mask2remove.size:
        clean = np.delete(X, sample_mask2remove, axis=1)
        sample_mask = np.ones((1, ns), dtype=bool)
        sample_mask[0, sample_mask2remove] = False
    else:
        clean = X
        sample_mask = np.ones((1, ns), dtype=bool)

    if show:
        import matplotlib.pyplot as plt
        f, ax = plt.subplots(nc, sharex=True, figsize=(8, 5))
        times = np.arange(ns) / float(sfreq)
        for i in range(nc):
            ax[i].fill_between(times, 0, 1, where=sample_mask.flat,
                               transform=ax[i].get_xaxis_transform(),
                               facecolor="none", hatch="...", edgecolor="k",
                               label="selected window")
            ax[i].plot(times, X[i], lw=.5, label="EEG")
            ax[i].set_ylim([-50, 50])
            # ax[i].set_ylabel(raw.ch_names[i])
            ax[i].set_yticks([])
        ax[i].set_xlabel("Time (s)")
        ax[i].set_ylabel(f"ch{i}")
        ax[0].legend(fontsize="small", bbox_to_anchor=(1.04, 1),
                     borderaxespad=0)
        plt.subplots_adjust(hspace=0, right=0.75)
        plt.suptitle("Clean windows")
        plt.show()

    return clean, sample_mask


def asr_calibrate(X, sfreq, cutoff=5, blocksize=100, win_len=0.5,
                  win_overlap=0.66, max_dropout_fraction=0.1,
                  min_clean_fraction=0.25, method="euclid", estimator="scm"):
    """Calibration function for the Artifact Subspace Reconstruction method.

    The input to this data is a multi-channel time series of calibration data.
    In typical uses the calibration data is clean resting EEG data of ca. 1
    minute duration (can also be longer). One can also use on-task data if the
    fraction of artifact content is below the breakdown point of the robust
    statistics used for estimation (50% theoretical, ~30% practical). If the
    data has a proportion of more than 30-50% artifacts then bad time windows
    should be removed beforehand. This data is used to estimate the thresholds
    that are used by the ASR processing function to identify and remove
    artifact components.

    The calibration data must have been recorded for the same cap design from
    which data for cleanup will be recorded, and ideally should be from the
    same session and same subject, but it is possible to reuse the calibration
    data from a previous session and montage to the extent that the cap is
    placed in the same location (where loss in accuracy is more or less
    proportional to the mismatch in cap placement).

    The calibration data should have been high-pass filtered (for example at
    0.5Hz or 1Hz using a Butterworth IIR filter).

    Parameters
    ----------
    X : array, shape=([n_trials, ]n_channels, n_samples)
        *zero-mean* (e.g., high-pass filtered) and reasonably clean EEG of not
        much less than 30 seconds (this method is typically used with 1 minute
        or more).
    sfreq : float
        Sampling rate of the data, in Hz.
    cutoff: float
        Standard deviation cutoff for rejection. X portions whose variance is
        larger than this threshold relative to the calibration data are
        considered missing data and will be removed. The most aggressive value
        that can be used without losing too much EEG is 2.5. A quite
        conservative value would be 5 (default=5).
    blocksize : int
        Block size for calculating the robust data covariance and thresholds,
        in samples; allows to reduce the memory and time requirements of the
        robust estimators by this factor (down to n_chans x n_chans x n_samples
        x 16 / blocksize bytes) (default=100).
    win_len : float
        Window length that is used to check the data for artifact content. This
        is ideally as long as the expected time scale of the artifacts but
        short enough to allow for several 1000 windows to compute statistics
        over (default=0.5).
    win_overlap : float
        Window overlap fraction. The fraction of two successive windows that
        overlaps. Higher overlap ensures that fewer artifact portions are going
        to be missed, but is slower (default=0.66).
    max_dropout_fraction : float
        Maximum fraction of windows that can be subject to signal dropouts
        (e.g., sensor unplugged), used for threshold estimation (default=0.1).
    min_clean_fraction : float
        Minimum fraction of windows that need to be clean, used for threshold
        estimation (default=0.25).
    method : {'euclid', 'riemann'}
        Metric to compute the covariance matrix average.

    Returns
    -------
    M : array
        Mixing matrix.
    T : array
        Threshold matrix.

    """
    logging.debug("[ASR] Calibrating...")

    # set number of channels and number of samples
    [nc, ns] = X.shape

    # filter the data
    X, _zf = yulewalk_filter(X, sfreq, ab=None)

    # window length for calculating thresholds
    N = int(np.round(win_len * sfreq))

    U = block_covariance(X, window=blocksize, overlap=win_overlap,
                         estimator=estimator)
    if method == "euclid":
        Uavg = geometric_median(U.reshape((-1, nc * nc)))
        Uavg = Uavg.reshape((nc, nc))
    else:  # method == 'riemann'
        Uavg = pyriemann.utils.mean.mean_covariance(U, metric="riemann")

    # get the mixing matrix M
    M = linalg.sqrtm(np.real(Uavg))
    D, Vtmp = linalg.eigh(M)
    # D, Vtmp = nonlinear_eigenspace(M, nc)  TODO
    V = Vtmp[:, np.argsort(D)]

    # get the threshold matrix T
    x = np.abs(np.dot(V.T, X))
    offsets = np.arange(0, ns - N, np.round(N * (1 - win_overlap))).astype(int)

    # go through all the channels and fit the EEG distribution
    mu = np.zeros(nc)
    sig = np.zeros(nc)
    for ichan in reversed(range(nc)):
        rms = x[ichan, :] ** 2
        Y = []
        for o in offsets:
            Y.append(np.sqrt(np.sum(rms[o:o + N]) / N))

        mu[ichan], sig[ichan], alpha, beta = fit_eeg_distribution(
            Y, min_clean_fraction, max_dropout_fraction)

    T = np.dot(np.diag(mu + cutoff * sig), V.T)
    logging.debug("[ASR] Calibration done.")
    return M, T


def asr_process(X, X_filt, state, cov=None, detrend=False, method="riemann",
                sample_weight=None):
    """Apply Artifact Subspace Reconstruction method.

    This function is used to clean multi-channel signal using the ASR method.
    The required inputs are the data matrix, the sampling rate of the data, and
    the filter state.

    Parameters
    ----------
    X : array, shape=([n_trials, ]n_channels, n_samples)
        Raw data.
    X_filt : array, shape=([n_trials, ]n_channels, n_samples)
        Yulewalk-filtered epochs to estimate covariance. Optional if covariance
        is provided.
    state : dict
        Initial ASR parameters (as derived by :func:`asr_calibrate`):

        - ``M`` : Mixing matrix
        - ``T`` : Threshold matrix
        - ``R`` : Previous reconstruction matrix (array | None)
    cov : array, shape=([n_trials, ]n_channels, n_channels) | None
        Covariance. If None (default), then it is computed from ``X_filt``. If
        a 3D array is provided, the average covariance is computed from all the
        elements in it.
    detrend : bool
        If True, detrend filtered data (default=False).
    method : {'euclid', 'riemann'}
        Metric to compute the covariance matric average.

    Returns
    -------
    clean : array, shape=([n_trials, ]n_channels, n_samples)
        Clean data.
    state : 3-tuple
        Output ASR parameters.

    """
    M, T, R = state.values()
    [nc, ns] = X.shape

    if cov is None:
        if detrend:
            X_filt = signal.detrend(X_filt, axis=1, type="constant")
        cov = block_covariance(X_filt, window=nc ** 2)

    cov = cov.squeeze()
    if cov.ndim == 3:
        if method == "riemann":
            cov = pyriemann.utils.mean.mean_covariance(
                cov, metric="riemann", sample_weight=sample_weight)
        else:
            cov = geometric_median(cov.reshape((-1, nc * nc)))
            cov = cov.reshape((nc, nc))

    maxdims = int(np.fix(0.66 * nc))  # constant TODO make param

    # do a PCA to find potential artifacts
    if method == "riemann":
        D, Vtmp = nonlinear_eigenspace(cov, nc)  # TODO
    else:
        D, Vtmp = linalg.eigh(cov)

    V = np.real(Vtmp[:, np.argsort(D)])
    D = np.real(D[np.argsort(D)])

    # determine which components to keep (variance below directional threshold
    # or not admissible for rejection)
    keep = (D < np.sum(np.dot(T, V)**2, axis=0))
    keep += (np.arange(nc) < nc - maxdims)

    # update the reconstruction matrix R (reconstruct artifact components using
    # the mixing matrix)
    if keep.all():
        R = np.eye(nc)  # trivial case
    else:
        VT = np.dot(V.T, M)
        demux = VT * keep[:, None]
        R = np.dot(np.dot(M, linalg.pinv(demux)), V.T)

    if state["R"] is not None:
        # apply the reconstruction to intermediate samples (using raised-cosine
        # blending)
        blend = (1 - np.cos(np.pi * np.arange(ns) / ns)) / 2
        clean = blend * R.dot(X) + (1 - blend) * state["R"].dot(X)
    else:
        clean = R.dot(X)

    state["R"] = R

    return clean, state



"""Tools for working with auditory signals."""
import numpy as np
import scipy.signal as ss


def erb_bandwidth(fc):
    """Bandwidth of an Equivalent Rectangular Bandwidth (ERB).

    Parameters
    ----------
    fc : ndarray
        Center frequency, or center frequencies, of the filter.

    Returns
    -------
    ndarray or float
        Equivalent rectangular bandwidth of the filter(s).

    """
    # In Hz, according to Glasberg and Moore (1990)
    return 24.7 + fc / 9.265


def erb2hz(erb):
    """Convert ERB-rate values to the corresponding frequencies in Hz."""
    f = (1. / 0.00437) * np.sign(erb) * (np.exp(np.abs(erb) / 9.2645) - 1)
    return f


def hz2erb(f):
    """Convert frequencies to the corresponding ERB-rates.

    Notes
    -----
    There is a round-off error in the Glasberg & Moore paper, as 1000 / (24.7 *
    4.37) * log(10) = 21.332 and not 21.4 as is stated.

    """
    # erb = 21.332 * np.sign(f) * np.log10(1 + np.abs(f) * 0.00437)
    erb = 9.2645 * np.sign(f) * np.log(1 + np.abs(f) * 0.00437)
    return erb


def erbspace(flow, fhigh, n):
    """Generate n equidistantly spaced points on ERB scale."""
    audlimits = hz2erb([flow, fhigh])
    y = erb2hz(np.linspace(audlimits[0], audlimits[1], n))

    bw = (audlimits[1] - audlimits[0]) / (n - 1)

    # Set the endpoints to be exactly what the user specified, instead of the
    # calculated values
    y[0] = flow
    y[-1] = fhigh

    return y, bw


class GammatoneFilterbank:
    """Gammatone Filterbank.

    This class computes the filter coefficients for a bank of Gammatone
    filters. These filters were defined by Patterson and Holdworth for
    simulating the cochlea, and originally implemented in [1]_.

    Parameters
    ----------
    sfreq : float
        Sampling frequency of the signals to filter.
    cf : array_like
        Center frequencies of the filterbank.
    b : float
        beta of the gammatone filters (default=1.019).
    order : int
        Order (default=1).
    q : float
        Q-value of the ERB (default=9.26449).
    min_bw : float
        Minimum bandwidth of an ERB.

    Notes
    -----
    The python was adapted from Alexandre Chabot-Leclerc's pambox, and Jason
    Heeris' gammatone toolbox:
    https://github.com/achabotl/pambox
    https://github.com/detly/gammatone

    References
    ----------
    .. [1] Slaney, M. (1993). An efficient implementation of the
       Patterson-Holdsworth auditory filter bank. Apple Computer, Perception
       Group, Tech. Rep, 35(8).

    """

    def __init__(self, sfreq, cf, b=1.019, order=1, q=9.26449, min_bw=24.7):
        self.sfreq = sfreq
        try:
            len(cf)
        except TypeError:
            cf = [cf]
        self.cf = np.asarray(cf)
        self.b = b
        self.erb_order = order
        self.q = q
        self.min_bw = min_bw

    def _get_coefs(self):
        """Compute the filter coefficients for a bank of Gammatone filters.

        These filters were defined by Patterson and Holdworth for simulating
        the cochlea.
        """
        T = 1 / self.sfreq
        cf = self.cf
        q = self.q  # Glasberg and Moore Parameters
        width = 1
        min_bw = 24.7
        order = 1

        erb = width * ((cf / q)**order + min_bw**order)**(1 / order)
        B = 1.019 * 2 * np.pi * erb

        arg = 2 * cf * np.pi * T
        vec = np.exp(2j * arg)

        A0 = T
        A2 = 0
        B0 = 1
        B1 = -2 * np.cos(arg) / np.exp(B * T)
        B2 = np.exp(-2 * B * T)

        rt_pos = np.sqrt(3 + 2**1.5)
        rt_neg = np.sqrt(3 - 2**1.5)

        common = -T * np.exp(-(B * T))

        # TODO: This could be simplified to a matrix calculation involving the
        # constant first term and the alternating rt_pos/rt_neg and +/-1 second
        # terms
        k11 = np.cos(arg) + rt_pos * np.sin(arg)
        k12 = np.cos(arg) - rt_pos * np.sin(arg)
        k13 = np.cos(arg) + rt_neg * np.sin(arg)
        k14 = np.cos(arg) - rt_neg * np.sin(arg)

        A11 = common * k11
        A12 = common * k12
        A13 = common * k13
        A14 = common * k14

        gain_arg = np.exp(1j * arg - B * T)

        gain = np.abs((vec - gain_arg * k11) *
                      (vec - gain_arg * k12) *
                      (vec - gain_arg * k13) *
                      (vec - gain_arg * k14) *
                      (T * np.exp(B * T) /
                       (-1 / np.exp(B * T) + 1 + vec *
                        (1 - np.exp(B * T)))) ** 4)

        allfilts = np.ones_like(cf)

        fcoefs = np.column_stack([
            A0 * allfilts, A11, A12, A13, A14, A2 * allfilts,
            B0 * allfilts, B1, B2,
            gain
        ])

        return fcoefs

    def filter(self, X):
        """Filter X along its last dimension.

        Process an input waveform with a gammatone filter bank. This function
        takes a single sound vector, and returns an array of filter outputs,
        one channel per row.

        Parameters
        ----------
        X : ndarray, shape=(n_chans, n_times)
            Signal to filter.

        Returns
        -------
        ndarray
            Filtered signals with shape ``(M, N)``, where ``M`` is the number
            of channels, and ``N`` is the input signal's number of samples.
        """
        coefs = self._get_coefs()
        output = np.zeros((coefs[:, 9].shape[0], X.shape[0]))

        gain = coefs[:, 9]
        # A0, A11, A2
        As1 = coefs[:, (0, 1, 5)]
        # A0, A12, A2
        As2 = coefs[:, (0, 2, 5)]
        # A0, A13, A2
        As3 = coefs[:, (0, 3, 5)]
        # A0, A14, A2
        As4 = coefs[:, (0, 4, 5)]
        # B0, B1, B2
        Bs = coefs[:, 6:9]

        # Loop over channels
        for idx in range(0, coefs.shape[0]):
            # These seem to be reversed (in the sense of A/B order), but that's
            # what the original code did... Replacing these with polynomial
            # multiplications reduces both accuracy and speed.
            y1 = ss.lfilter(As1[idx], Bs[idx], X)
            y2 = ss.lfilter(As2[idx], Bs[idx], y1)
            y3 = ss.lfilter(As3[idx], Bs[idx], y2)
            y4 = ss.lfilter(As4[idx], Bs[idx], y3)
            output[idx, :] = y4 / gain[idx]

        return output


class AuditoryFilterbank(GammatoneFilterbank):
    """Special case of Gammatone filterbank with preset center frequencies."""

    def __init__(self, sfreq, b=1.019, order=1, q=9.26449, min_bw=24.7):

        cf = np.asarray([63, 80, 100, 125, 160, 200, 250, 315, 400, 500,
                         30, 800, 1000, 1250, 1600, 2000, 2500, 3150, 4000,
                         5000, 6300, 8000])

        super().__init__(
            sfreq=sfreq, cf=cf, b=b, order=order, q=q, min_bw=min_bw)



"""Test robust detrending."""
import numpy as np
from scipy.signal import butter, lfilter

from meegkit.detrend import create_masked_weight, detrend, reduce_ringing, regress

rng = np.random.default_rng(9)

def test_regress():
    """Test regression."""
    # Simple regression example, no weights
    # fit random walk
    y = np.cumsum(rng.standard_normal((1000, 1)), axis=0)
    x = np.arange(1000.)[:, None]
    x = np.hstack([x, x ** 2, x ** 3])
    [b, z] = regress(y, x)

    # Simple regression example, with weights
    y = np.cumsum(rng.standard_normal((1000, 1)), axis=0)
    w = rng.random(y.shape)
    [b, z] = regress(y, x, w)

    # Downweight 1st half of the data
    y = np.cumsum(rng.standard_normal((1000, 1)), axis=0) + 1000
    w = np.ones(y.shape[0])
    w[:500] = 0
    [b, z] = regress(y, x, w)

    # # Multichannel regression
    y = np.cumsum(rng.standard_normal((1000, 2)), axis=0)
    w = np.ones(y.shape[0])
    [b, z] = regress(y, x, w)
    assert z.shape == (1000, 2)
    assert b.shape == (2, 1)

    # Multichannel regression
    y = np.cumsum(rng.standard_normal((1000, 2)), axis=0)
    w = np.ones(y.shape)
    w[:, 1] = .8
    [b, z] = regress(y, x, w)
    assert z.shape == (1000, 2)
    assert b.shape == (2, 3)


def test_detrend(show=False):
    """Test detrending."""
    # basic
    x = np.arange(100)[:, None]  # trend
    source = rng.standard_normal(x.shape)
    x = x + source
    y, _, _ = detrend(x, 1)

    assert y.shape == x.shape

    # detrend biased random walk
    x = np.cumsum(rng.standard_normal((1000, 1)) + 0.1)
    y, _, _ = detrend(x, 3)

    assert y.shape == x.shape

    # test weights
    trend = np.linspace(0, 100, 1000)[:, None]
    data = 3 * rng.standard_normal(trend.shape)
    data[:100, :] = 100
    x = trend + data
    w = np.ones(x.shape)
    w[:100, :] = 0

    # Without weights – detrending fails
    y, _, _ = detrend(x, 3, None, show=show)

    # With weights – detrending works
    yy, _, _ = detrend(x, 3, w, show=show)

    assert y.shape == x.shape
    assert yy.shape == x.shape
    assert np.all(np.abs(yy[100:] - data[100:]) < 1.)

    # detrend higher-dimensional data
    x = np.cumsum(rng.standard_normal((1000, 16)) + 0.1, axis=0)
    y, _, _ = detrend(x, 1, show=False)

    # detrend higher-dimensional data with order 3 polynomial
    x = np.cumsum(rng.standard_normal((1000, 16)) + 0.1, axis=0)
    y, _, _ = detrend(x, 3, basis="polynomials", show=True)

    # detrend with sinusoids
    x = rng.standard_normal((1000, 2))
    x += 2 * np.sin(2 * np.pi * np.arange(1000) / 200)[:, None]
    y, _, _ = detrend(x, 5, basis="sinusoids", show=True)

    # trial-masked detrending
    trend = np.linspace(0, 100, 1000)[:, None]
    data = 3 * rng.standard_normal(trend.shape)
    data[:100, :] = 100
    x = trend + data
    events = np.arange(30, 970, 40)
    tmin, tmax, sfreq = -0.2, 0.3, 20
    w = create_masked_weight(x, events, tmin, tmax, sfreq)
    y, _, _ = detrend(x, 1, w, basis="polynomials", show=show)


def test_ringing():
    """Test reduce_ringing function."""
    x = np.arange(1000) < 1
    [b, a] = butter(6, 0.2)     # Butterworth filter design
    x = lfilter(b, a, x) * 50    # Filter data using above filter
    x = np.roll(x, 500)
    signal = rng.standard_normal((1000, 2))
    x = x[:, None] + signal
    reduce_ringing(x, samples=np.array([500]))

    # np.testing.assert_array_almost_equal(y, signal, 2)

if __name__ == "__main__":
    import pytest
    pytest.main([__file__])
    # test_detrend(False)



"""Local Outlier Factor (LOF)."""
# Authors: Velu Prabhakar Kumaravel <vkumaravel@fbk.eu>
# License: BSD-3-Clause

import logging

from sklearn.neighbors import LocalOutlierFactor


class LOF:
    """Local Outlier Factor.

    Local Outlier Factor (LOF) is an automatic, density-based outlier detection
    algorithm based on [1]_ and [2]_.

    Parameters
    ----------
    n_neighbours : int
        Number of neighbours defining the local neighbourhood.
    metric: str in {'euclidean', 'nan_euclidean', 'cosine',
                    'cityblock', 'manhattan'}
        Metric to use for distance computation. Default is “euclidean”
    threshold : float
        Threshold to define outliers. Theoretical threshold ranges anywhere
        between 1.0 and any integer. Default: 1.5

    Notes
    -----
    It is recommended to perform a CV (e.g., 10-fold) on training set to
    calibrate this parameter for the given M/EEG dataset.

    See [2]_ for details.

    References
    ----------
    .. [1] Breunig M, Kriegel HP, Ng RT, Sander J.
        2000. LOF: identifying density-based local outliers.
        SIGMOD Rec. 29, 2, 93-104. https://doi.org/10.1145/335191.335388
    .. [2] Kumaravel VP, Buiatti M, Parise E, Farella E.
        2022. Adaptable and Robust EEG Bad Channel Detection Using
        Local Outlier Factor (LOF). Sensors (Basel). 2022 Sep 27;22(19):7314.
        doi: 10.3390/s22197314. PMID: 36236413; PMCID: PMC9571252.

    """

    def __init__(self, n_neighbors=20,  metric="euclidean",
                 threshold=1.5, **kwargs):

        self.n_neighbors = n_neighbors
        self.metric = metric
        self.threshold = threshold

    def predict(self, X):
        """Detect bad channels using Local Outlier Factor algorithm.

        Parameters
        ----------
        X : array, shape=(n_channels, n_samples)
            The data X should have been high-pass filtered.

        Returns
        -------
        bad_channel_indices : Detected bad channel indices.

        """
        if X.ndim == 3:  # in case the input data is epoched
            logging.warning("Expected input data with shape "
                            "(n_channels, n_samples)")
            return []

        if self.n_neighbors >= X.shape[0]:
            logging.warning("Number of neighbours cannot be greater than the "
                            "number of channels")
            return []

        if self.threshold < 1.0:
            logging.warning("Invalid threshold. Try a positive integer >= 1.0")
            return []

        clf = LocalOutlierFactor(self.n_neighbors)
        logging.debug("[LOF] Predicting bad channels")
        clf.fit_predict(X)
        lof_scores = clf.negative_outlier_factor_
        bad_channel_indices = -lof_scores >= self.threshold

        return bad_channel_indices



import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from meegkit.utils import (
    bootstrap_ci,
    cronbach,
    demean,
    find_outlier_samples,
    find_outlier_trials,
    fold,
    mean_over_trials,
    multishift,
    multismooth,
    relshift,
    rms,
    robust_mean,
    shift,
    shiftnd,
    unfold,
    widen_mask,
)

rng = np.random.default_rng(9)


def _sim_data(n_times, n_chans, n_trials, noise_dim, SNR=1, t0=100):
    """Create synthetic data."""
    # source
    source = np.sin(2 * np.pi * np.linspace(0, .5, n_times - t0))[np.newaxis].T
    s = source * rng.standard_normal((1, n_chans))
    s = s[:, :, np.newaxis]
    s = np.tile(s, (1, 1, n_trials))
    signal = np.zeros((n_times, n_chans, n_trials))
    signal[t0:, :, :] = s

    # noise
    noise = np.dot(
        unfold(rng.standard_normal((n_times, noise_dim, n_trials))),
        rng.standard_normal((noise_dim, n_chans)))
    noise = fold(noise, n_times)

    # mix signal and noise
    signal = SNR * signal / rms(signal.flatten())
    noise = noise / rms(noise.flatten())
    noisy_data = signal + noise
    return noisy_data, signal


def test_multishift():
    """Test matrix multi-shifting."""
    # multishift()
    x = np.array([
        [1, 2, 3, 4],
        [5, 6, 7, 8]])
    assert_equal(multishift(x, [1], axis=0), np.array([[0, 0, 0, 0],
                                                       [1, 2, 3, 4]]))

    # multishift 3d, solution='valid'
    x = np.ones((4, 4, 3))
    x[..., 1] *= 2
    x[..., 2] *= 3
    xx = multishift(x, [-1, -2], reshape=True, solution="valid")
    assert_equal(xx[..., 0], np.array([[1., 1., 1., 1., 1., 1., 1., 1.],
                                       [1., 1., 1., 1., 1., 1., 1., 1.]]))
    assert_equal(xx[..., 1], np.array([[1., 1., 1., 1., 1., 1., 1., 1.],
                                       [1., 1., 1., 1., 1., 1., 1., 1.]]) * 2)

    # relshift() 1d
    y, y_ref = relshift([1, 2, 3, 4], [11, 12, 13, 14], [1], axis=0)
    assert_equal(y.flatten(), [0, 1, 2, 3])
    assert_equal(y_ref.flatten(), [0, 12, 13, 14])

    y, y_ref = relshift(np.arange(1, 6), np.arange(1, 6), [0, 1], axis=0)
    assert_equal(y, np.array([[1., 0.],
                              [2., 1.],
                              [3., 2.],
                              [4., 3.],
                              [5., 4.]]))

    # relshift() 2d
    x = [[1, 2, 3, 4],
         [5, 6, 7, 8]]
    x_ref = [[11, 12, 13, 14],
             [15, 16, 17, 18]]

    y, y_ref = relshift(x, x_ref, [0, 1])
    assert_equal(y[..., 0], [[1, 2, 3, 4],
                             [5, 6, 7, 8]])
    assert_equal(y[..., 1], [[0, 0, 0, 0],
                             [1, 2, 3, 4]])
    assert_equal(y_ref[..., 0], [[11, 12, 13, 14],
                                 [15, 16, 17, 18]])
    assert_equal(y_ref[..., 1], [[0, 0, 0, 0],
                                 [15, 16, 17, 18]])


def test_shift():
    """Test matrix shifting."""
    x = np.arange(10)

    assert_equal(shiftnd(x, 2), np.array([0, 0, 0, 1, 2, 3, 4, 5, 6, 7]))

    # x2 = array([[0, 1, 2, 3, 4],
    #             [5, 6, 7, 8, 9]])
    x2 = np.reshape(x, (2, 5))

    # Test shifts along several dimensions and directions
    assert_equal(shiftnd(x2, 1), np.array([[0, 0, 1, 2, 3],
                                           [4, 5, 6, 7, 8]]))
    assert_equal(shiftnd(x2, -2), np.array([[2, 3, 4, 5, 6],
                                            [7, 8, 9, 0, 0]]))
    assert_equal(shiftnd(x2, 1, axis=0), np.array([[0, 0, 0, 0, 0],
                                                   [0, 1, 2, 3, 4]]))
    assert_equal(shiftnd(x2, -1, axis=0), np.array([[5, 6, 7, 8, 9],
                                                    [0, 0, 0, 0, 0]]))
    assert_equal(shiftnd(x2, 1, axis=1), np.array([[0, 0, 1, 2, 3],
                                                   [0, 5, 6, 7, 8]]))

    # Same for `shift`
    assert_equal(shift(x2, 1, axis=0), np.array([[0, 0, 0, 0, 0],
                                                 [0, 1, 2, 3, 4]]))
    assert_equal(shift(x2, -1, axis=0), np.array([[5, 6, 7, 8, 9],
                                                  [0, 0, 0, 0, 0]]))
    assert_equal(shift(x2, 1, axis=1), np.array([[0, 0, 1, 2, 3],
                                                 [0, 5, 6, 7, 8]]))


def test_widen_mask():
    """Test binary mask operations."""
    test = np.array([0, 0, 0, 1, 0, 0, 0])

    # test 1d
    assert_equal(widen_mask(test, -2), [0, 1, 1, 1, 0, 0, 0])
    assert_equal(widen_mask(test, 2), [0, 0, 0, 1, 1, 1, 0])

    # test nd
    assert_equal(widen_mask(test[None, :], -2, axis=1),
                 [[0, 1, 1, 1, 0, 0, 0], ])
    assert_equal(widen_mask(test[None, None, :], -2, axis=2),
                 [[[0, 1, 1, 1, 0, 0, 0], ], ])


def test_multismooth():
    """Test smoothing."""
    x = (rng.standard_normal((1000, 1)) / 2 +
         np.cos(2 * np.pi * 3 * np.linspace(0, 20, 1000))[:, None])

    for i in np.arange(1, 10, 1):
        y = multismooth(x, i)
        assert x.shape == y.shape

    y = multismooth(x, np.arange(5) + 1)
    assert y.shape == x.shape + (5, )


def test_demean(show=False):
    """Test demean."""
    n_trials = 100
    n_chans = 8
    n_times = 1000
    x = rng.standard_normal((n_times, n_chans, n_trials))
    x, s = _sim_data(n_times, n_chans, n_trials, 8, SNR=10)

    # 1. demean and check trial average is almost zero
    x1 = demean(x)
    assert_almost_equal(x1.mean(2).mean(0), np.zeros((n_chans,)))
    assert_almost_equal(x1.mean(2), mean_over_trials(x1)[0])

    # 2. now use weights : baseline = first 100 samples
    times = np.arange(n_times)
    weights = np.zeros_like(times)
    weights[:100] = 1
    x2 = demean(x, weights)

    if show:
        import matplotlib.pyplot as plt
        f, ax = plt.subplots(3, 1)
        ax[0].plot(times, x[:, 0].mean(-1), label="noisy_data")
        ax[0].plot(times, s[:, 0].mean(-1), label="signal")
        ax[0].legend()
        ax[1].plot(times, x1[:, 0].mean(-1), label="mean over entire epoch")
        ax[1].legend()
        ax[2].plot(x2[:, 0].mean(-1), label="weighted mean")
        plt.gca().set_prop_cycle(None)
        ax[2].plot(s[:, 0].mean(-1), "k:")
        ax[2].legend()
        plt.show()

    # assert mean is ~= 0 during baseline
    assert_almost_equal(x2[:100].mean(2).mean(0), np.zeros((n_chans,)))

    # assert trial average is close to source signal
    assert np.all(x2.mean(2) - s.mean(2) < 1)

    # 3. try different shape weights (these should all be equivalent)
    weights = np.zeros((len(times), 1, 1))
    weights[:100] = 1
    x1 = demean(x, weights)

    weights = np.zeros((len(times), n_chans, 1))
    weights[:100] = 1
    x2 = demean(x, weights)

    weights = np.zeros((len(times), n_chans, n_trials))
    weights[:100] = 1
    x3 = demean(x, weights)

    np.testing.assert_array_equal(x1, x2)
    np.testing.assert_array_equal(x1, x3)


def test_computeci():
    """Compute CI."""
    x = rng.standard_normal((1000, 8, 100))
    ci_low, ci_high = bootstrap_ci(x)

    assert ci_low.shape == (1000, 8)
    assert ci_high.shape == (1000, 8)

    # TODO axis as tuple
    # ci_low, ci_high = bootstrap_ci(x, axis=(1, 2))
    # assert ci_low.shape == (1000,)
    # assert ci_high.shape == (1000,)

    x = rng.standard_normal((1000, 100))
    ci_low, ci_high = bootstrap_ci(x)

    assert ci_low.shape == (1000,)
    assert ci_high.shape == (1000,)


def test_outliers(show=False):
    """Test outlier detection."""
    x = rng.standard_normal((250, 8, 50))  # 50 trials, 8, channels
    x[..., :5] *= 10  # 5 first trials are outliers

    # Pass standard threshold
    idx, _ = find_outlier_trials(x, 2, show=show)
    np.testing.assert_array_equal(idx, np.arange(5))

    idx, _ = find_outlier_trials(x, [2, 2], show=True)
    np.testing.assert_array_equal(idx, np.arange(5))

    idx = find_outlier_samples(x, 5)
    assert idx.shape == x.shape

    idx = find_outlier_samples(x, 5, 10)
    assert idx.shape == x.shape


def test_cronbach():
    """Test Cronbach's alpha."""
    np.random.default_rng(9)

    X, _ = _sim_data(800, 8, 80, noise_dim=6, SNR=.2)
    X = X.transpose([2, 1, 0])  # trials, channels, samples
    alpha, lo, hi = cronbach(X, tmin=0, n_bootstrap=200)
    print(alpha)
    assert np.all(lo < hi)

    X, _ = _sim_data(800, 8, 80, noise_dim=6, SNR=2)
    X = X.transpose([2, 1, 0])
    alpha2, lo, hi = cronbach(X, tmin=100, n_bootstrap=200)
    print(alpha2)
    assert np.sum(alpha2 > alpha) >= 6

    m = robust_mean(X, axis=0)
    assert m.shape == (X.shape[1], X.shape[2])

if __name__ == "__main__":
    import pytest
    pytest.main([__file__])

    # test_outliers()
    # import matplotlib.pyplot as plt
    # x = rng.standard_normal((1000,)
    # y = multismooth(x, np.arange(1, 200, 4))
    # plt.imshow(y.T, aspect='auto')
    # plt.show()

    # test_cronbach()



"""Test signal utils."""
import numpy as np
from scipy.signal import butter, freqz, lfilter

from meegkit.utils.sig import stmcb, teager_kaiser

rng = np.random.default_rng(9)

def test_teager_kaiser(show=False):
    """Test Teager-Kaiser Energy."""
    x = 2 * rng.random((1000, 2)) - 1
    x[100, 0] = 5
    x[200, 0] = 5
    x += np.cumsum(rng.standard_normal((1000, 1)) + 0.1, axis=0) / 1000
    for i in range(1, 5):
        print(i)
        y = teager_kaiser(x, M=i, m=1)
        assert y.shape[0] == 1000 - 2 * i

    if show:
        import matplotlib.pyplot as plt
        plt.figure()
        # plt.plot((x[1:, 0] - x[1:, 0].mean()) / np.nanstd(x[1:, 0]))
        # plt.plot((y[..., 0] - y[..., 0].mean()) / np.nanstd(y[..., 0]))
        plt.plot(x[1:, 0], label="X")
        plt.plot(y[..., 0], label="Y")
        plt.legend()
        plt.show()


def test_stcmb(show=True):
    """Test stcmb."""
    x = np.arange(100) < 1
    [b, a] = butter(6, 0.2)     # Butterworth filter design
    y = lfilter(b, a, x)        # Filter data using above filter

    w, h = freqz(b, a, 128)     # Frequency response
    [bb, aa] = stmcb(y, u_in=None, q=4, p=4, niter=5)
    ww, hh = freqz(bb, aa, 128)

    if show:
        import matplotlib.pyplot as plt
        f, ax = plt.subplots(2, 1)
        ax[0].plot(x, label="step")
        ax[0].plot(y, label="filt")
        ax[1].plot(w, np.abs(h), label="real")
        ax[1].plot(ww, np.abs(hh), label="stcmb")
        ax[0].legend()
        ax[1].legend()
        plt.show()

    np.testing.assert_allclose(h, hh, rtol=2)  # equal to 2%

if __name__ == "__main__":
    test_teager_kaiser()
    test_stcmb()



"""ASR test."""
import os

import matplotlib.pyplot as plt
import numpy as np
import pytest
from scipy import signal

from meegkit.asr import ASR, asr_calibrate, asr_process, clean_windows
from meegkit.utils.asr import yulewalk, yulewalk_filter
from meegkit.utils.matrix import sliding_window

# Data files
THIS_FOLDER = os.path.dirname(os.path.abspath(__file__))
# file = os.path.join(THIS_FOLDER, 'data', 'eeg_raw.fif')
# raw = mne.io.read_raw_fif(file, preload=True)
# raw.filter(.5, 30)
# raw._data *= 1e6  # to uV for easy plotting
# raw.crop(0, 60)  # keep 60s only
# raw.pick_types(eeg=True, misc=False)
# raw = raw._data
rng = np.random.default_rng(9)


@pytest.mark.parametrize(argnames="sfreq", argvalues=(125, 250, 256, 2048))
def test_yulewalk(sfreq, show=False):
    """Test that my version of yulewelk works just like MATLAB's."""
    # Temp fix, values are computed in matlab using yulewalk.m
    if sfreq == 125:
        a = [1, -0.983952187817050, -0.520232502560362, 0.603540557711479,
             0.116893105621457, -0.0291261609247754, -0.282359853603720,
             0.0407847933579206, 0.103437108246108]
        b = [1.08742316795540, -1.83643555381637, 0.573976014496824,
             0.361020603610170, 0.0592714561864745, 0.0767631759850725,
             -0.498304757808424, 0.276872948140515, -0.00693079202803615]
    elif sfreq == 256:
        a = [1, -1.70080396393018, 1.92328303910588, -2.08269297269299,
             1.59826387425574, -1.07358541839301, 0.567971922565269,
             -0.188618149976820, 0.0572954115997260]
        b = [1.75870131417701, -4.32676243944586, 5.79998800310163,
             -6.23966254635480, 5.37680790468827, -3.79382188933751,
             2.16491080952266, -0.859139256986372, 0.256936112562797]
    elif sfreq == 250:
        a = [1, -1.63849492766660, 1.73987814299054, -1.83638657883455,
             1.39241775367979, -0.953780426622192, 0.505158779550744,
             -0.159504514603054, 0.0545278399847976]
        b = [1.73133310854258, -4.16813353295698, 5.37379900844170,
             -5.57212564343883, 4.70122651316511, -3.34208799655244,
             1.95045488724907, -0.766909658912065, 0.233281060974834]
    elif sfreq == 2048:
        a = [1, -7.37108717906118, 23.9532262496612, -44.8116760275397,
             52.7784619594557, -40.0682205405753, 19.1457849272128,
             -5.26406859898898, 0.637581364205670]
        b = [2.84929120330035, -21.2941639596528, 70.1293865360529,
             -132.920238664871, 158.567177443427, -121.909488069062,
             58.9853908881204, -16.4212688404351, 2.01391570212326]
    else:
        raise AttributeError("Currently sfreq must be 250, 256 or 2048...")

    # Theoretical values
    w0, h0 = signal.freqz(b, a, sfreq)

    f = np.array([0, 2, 3, 13, 16, 40, np.min((80, sfreq / 2 - 1)), sfreq / 2])
    f *= 2. / sfreq
    m = np.array([3, 0.75, 0.33, 0.33, 1, 1, 3, 3])
    [b, a] = yulewalk(8, f, m)

    w1, h1 = signal.freqz(b, a, sfreq)

    if show:
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(w0 / np.pi, np.abs(h0), label="matlab")
        ax.plot(w1 / np.pi, np.abs(h1), ":", label="mine")
        ax.set_title("Filter frequency response")
        ax.set_xlabel("Frequency [radians / second]")
        ax.set_ylabel("Amplitude [dB]")
        ax.grid(which="both", axis="both")
        ax.legend()
        # plt.show()

    np.testing.assert_almost_equal(np.abs(h0), np.abs(h1), decimal=4)

    f = [0, .6, .6, 1]              # Frequency breakpoints
    m = [1., 1., 0, 0]              # Magnitude breakpoints
    b, a = yulewalk(8, f, m)        # Filter design using least-squares method
    w, h = signal.freqz(b, a, 250)  # Frequency response of filter

    if show:
        plt.figure()
        plt.plot(f, m, label="ideal")
        plt.plot(w / np.pi, np.abs(h), "--", label="yw designed")
        plt.legend()
        plt.title("Comparison of Frequency Response Magnitudes")
        plt.legend()
        plt.show()


@pytest.mark.parametrize(argnames="n_chans", argvalues=(4, 8, 12))
def test_yulewalk_filter(n_chans, show=False):
    """Test yulewalk filter."""
    raw = np.load(os.path.join(THIS_FOLDER, "data", "eeg_raw.npy"))
    sfreq = 250
    n_chan_orig = raw.shape[0]
    raw = rng.standard_normal((n_chans, n_chan_orig)) @ raw
    raw_filt, iirstate = yulewalk_filter(raw, sfreq)

    if show:
        f, ax = plt.subplots(n_chans, sharex=True, figsize=(8, 5))
        for i in range(n_chans):
            ax[i].plot(raw[i], lw=.5, label="before")
            ax[i].plot(raw_filt[i], label="after", lw=.5)
            ax[i].set_ylim([-50, 50])
            if i < n_chans - 1:
                ax[i].set_yticks([])
        ax[i].set_xlabel("Time (s)")
        ax[i].set_ylabel(f"ch{i}")
        ax[0].legend(fontsize="small", bbox_to_anchor=(1.04, 1),
                     borderaxespad=0)
        plt.subplots_adjust(hspace=0, right=0.75)
        plt.suptitle("Before/after filter")
        plt.show()


def test_asr_functions(show=False, method="riemann"):
    """Test ASR functions (offline use).

    Note: this will not be optimal since the filter parameters will be
    estimated only once and not updated online as is intended.

    """
    raw = np.load(os.path.join(THIS_FOLDER, "data", "eeg_raw.npy"))
    sfreq = 250
    raw_filt = raw.copy()
    raw_filt, iirstate = yulewalk_filter(raw_filt, sfreq)

    # Train on a clean portion of data
    train_idx = np.arange(5 * sfreq, 45 * sfreq, dtype=int)

    # Clean data of high amplitude artifacts
    clean, sample_mask = clean_windows(raw[:, train_idx], sfreq)
    assert clean.shape[1] < train_idx.size  # make sure we removed artefacts

    M, T = asr_calibrate(clean, sfreq, method=method, cutoff=2,
                         max_dropout_fraction=.2)
    state = dict(M=M, T=T, R=None)
    clean, _ = asr_process(raw, raw_filt, state, method=method)

    if show:
        f, ax = plt.subplots(8, sharex=True, figsize=(8, 5))
        for i in range(8):
            ax[i].fill_between(train_idx, 0, 1, color="grey", alpha=.3,
                               transform=ax[i].get_xaxis_transform(),
                               label="calibration window")
            ax[i].fill_between(train_idx, 0, 1, where=sample_mask.flat,
                               transform=ax[i].get_xaxis_transform(),
                               facecolor="none", hatch="...", edgecolor="k",
                               label="selected window")
            ax[i].plot(raw[i], lw=.5, label="before ASR")
            ax[i].plot(clean[i], label="after ASR", lw=.5)
            # ax[i].set_xlim([10, 50])
            ax[i].set_ylim([-50, 50])
            # ax[i].set_ylabel(raw.ch_names[i])
            if i < 7:
                ax[i].set_yticks([])
        ax[i].set_xlabel("Time (s)")
        ax[0].legend(fontsize="small", bbox_to_anchor=(1.04, 1),
                     borderaxespad=0)
        plt.subplots_adjust(hspace=0, right=0.75)
        plt.suptitle("Before/after ASR")
        plt.show()


@pytest.mark.parametrize(argnames="method", argvalues=("riemann", "euclid"))
@pytest.mark.parametrize(argnames="reref", argvalues=(False, True))
def test_asr_class(method, reref, show=False):
    """Test ASR class (simulate online use)."""
    raw = np.load(os.path.join(THIS_FOLDER, "data", "eeg_raw.npy"))
    sfreq = 250
    # Train on a clean portion of data
    train_idx = np.arange(5 * sfreq, 45 * sfreq, dtype=int)

    # Rereference
    if reref:
        # Rank deficient matrix
        raw2 = raw - np.nanmean(raw, axis=0, keepdims=True)
    else:
        raw2 = raw.copy()

    if reref:
        if method == "riemann":
            with pytest.raises(ValueError, match="Add regularization"):
                blah = ASR(method=method, estimator="scm")
                blah.fit(raw2[:, train_idx])

        asr = ASR(method=method, estimator="lwf")
        asr.fit(raw2[:, train_idx])
    else:
        asr = ASR(method=method, estimator="scm")
        asr.fit(raw2[:, train_idx])

    # Split into small windows
    X = sliding_window(raw2, window=int(sfreq // 2), step=int(sfreq // 2))
    X = X.swapaxes(0, 1)

    # Transform each trial
    Y = np.zeros_like(X)
    for i in range(X.shape[0]):
        Y[i] = asr.transform(X[i])

    # Transform all trials at once
    asr.reset()
    asr.fit(raw2[:, train_idx])
    Y2 = asr.transform(X)

    X = X.swapaxes(0, 1).reshape(8, -1)
    Y = Y.swapaxes(0, 1).reshape(8, -1)
    Y2 = Y2.swapaxes(0, 1).reshape(8, -1)
    times = np.arange(X.shape[-1]) / sfreq
    if show:
        f, ax = plt.subplots(8, sharex=True, figsize=(8, 5))
        for i in range(8):
            ax[i].plot(times, X[i], lw=.5, label="before ASR")
            ax[i].plot(times, Y[i], label="after ASR", lw=.5)
            ax[i].set_ylim([-50, 50])
            ax[i].set_ylabel(f"ch{i}")
            if i < 7:
                ax[i].set_yticks([])
        ax[i].set_xlabel("Time (s)")
        ax[0].legend(fontsize="small", bbox_to_anchor=(1.04, 1),
                     borderaxespad=0)
        plt.subplots_adjust(hspace=0, right=0.75)
        plt.suptitle("Before/after ASR")

        f, ax = plt.subplots(8, sharex=True, figsize=(8, 5))
        for i in range(8):
            ax[i].plot(times, Y[i], label="incremental", lw=.5)
            ax[i].plot(times, Y2[i], label="bulk", lw=.5)
            ax[i].plot(times, Y[i] - Y2[i], label="difference", lw=.5)
            if i < 7:
                ax[i].set_yticks([])
        ax[i].set_xlabel("Time (s)")
        plt.suptitle("incremental vs. bulk difference ")
        plt.show()

    # TODO: the transform() process is stochastic, so Y and Y2 are not going to
    # be entirely identical but close enough
    assert np.all(np.abs(Y - Y2) < 6), np.max(np.abs(Y - Y2))  # < 6uV diff
    assert np.all(np.isreal(Y)), "output should be real-valued"
    assert np.all(np.isreal(Y2)), "output should be real-valued"

    # Test different sampling rates
    with pytest.raises(ValueError):
        ASR(sfreq=60)

    ASR(sfreq=80)
    ASR(sfreq=100)
    ASR(sfreq=125)
    ASR(Sfreq=150)

if __name__ == "__main__":
    pytest.main([__file__])
    # test_yulewalk(250, True)
    # test_asr_functions(True)
    # test_asr_class(method='riemann', reref=True, show=False)
    # test_yulewalk_filter(16, True)



"""Time-shift PCA."""
import numpy as np

from .utils import (
    demean,
    fold,
    multishift,
    normcol,
    pca,
    regcov,
    theshapeof,
    tscov,
    tsxcov,
    unfold,
    unsqueeze,
)
from .utils.matrix import _check_shifts, _check_weights


def tspca(X, shifts=None, keep=None, threshold=None, weights=None,
          demean=False):
    """Time-shift PCA.

    Parameters
    ----------
    X : array, shape=(n_times, n_chans[, n_trials])
        Data array.
    shifts: array, shape=(n_shifts,)
        Array of shifts to apply.
    keep: int
        Number of components (shifted regressor PCs) to keep (default=all).
    threshold:
        Discard PCs with eigenvalues below this (default=1e-6).
    weights : array
        Sample weights.
    demean : bool
        If True, Epochs are centered before comuting PCA (default=0).

    Returns
    -------
    comps : array
        Principal components array.
    V : array, shape=(n_chans, n_comps)
        PCA weights.
    idx : array
        `X[idx]` maps to principal comps.

    """
    shifts, n_shifts = _check_shifts(shifts)
    weights = _check_weights(weights, X)

    n_samples, n_chans, n_trials = theshapeof(X)

    # offset of components relative to data
    offset = np.max((0, -np.min(shifts)))
    shifts += offset
    idx = offset + (np.arange(n_samples - np.max(shifts)))

    # remove mean
    if demean:
        X = unfold(X)
        X = demean(X, weights)
        X = fold(X, n_samples)

    # covariance
    C = tscov(X, shifts, weights)[0]

    # PCA matrix
    V, _ = pca(C, max_comps=keep, thresh=threshold)

    # apply PCA matrix to time-shifted data
    comps = np.zeros((np.size(idx), V.shape[1], n_trials))

    for t in np.arange(n_trials):
        comps[:, :, t] = np.dot(
            np.squeeze(multishift(X[:, :, t], shifts)),
            np.squeeze(V))

    return comps, V, idx


def tsr(X, R, shifts=None, wX=None, wR=None, keep=None, thresh=1e-12):
    """Time-shift regression.

    The basic idea is to project the signal `X` on a basis formed by the
    orthogonalized time-shifted `R`, and remove the projection. Supposing `R`
    gives a good observation of the noise that contaminates `X`, the noise is
    removed. By allowing time shifts, the algorithm finds the optimal FIR
    filter to apply to `R` so as to compensate for any convolutional mismatch
    between `X` and `R`.

    Parameters
    ----------
    X : array, shape=(n_samples, n_chans[, n_trials])
        Data to denoise.
    R : array, shape=(n_samples, n_comps[, n_trials])
        Reference data.
    shifts : array, shape=(n_shifts,)
        Array of shifts to apply to R (default=[0]).
    wX : array, shape=(n_samples, 1, n_trials)
        Weights to apply to `X`.
    wR : array, shape=(n_samples, 1, n_trials)
        Weights to apply to `R`.
    keep : int | None
        Number of shifted-R PCs to retain (default=all).
    thresh : float
        Ignore shifted-R PCs smaller than thresh (default=1e-12).

    Returns
    -------
    y : array
        Denoised data.
    idx : array
        X[idx] is aligned with `y`.
    mean : array
        Channel means (removed by TSR).
    weights : array
        Weights applied by TSR.

    """
    ndims = X.ndim
    X = unsqueeze(X)
    R = unsqueeze(R)
    shifts, n_shifts = _check_shifts(shifts)
    wX = _check_weights(wX, X)
    wR = _check_weights(wR, R)

    # adjust to make shifts non-negative
    initial_samples = X.shape[0]

    offset1 = np.max((0, -np.min(shifts)))
    idx = np.arange(offset1, X.shape[0])
    # X = X[idx, ...]
    # if len(wX) > 0:
    #     wX = wX[idx, ...]
    # if len(wR) > 0:
    #     wR = wR[:-offset1, ...]

    shifts = shifts + offset1  # shifts are now positive

    # adjust size of X
    offset2 = np.max((0, np.max(shifts)))
    # idx = np.arange(X.shape[0]) - offset2
    # idx = idx[idx >= 0]
    # X = X[idx, ...]
    # if len(wX) > 0:
    #     wX = wX[idx, ...]

    n_samples_X, n_chans_X, n_trials_X = theshapeof(X)
    n_samples_R, n_chans_R, n_trials_R = theshapeof(R)

    # consolidate weights into single weight matrix
    weights = np.zeros((n_samples_X, 1, n_trials_R))
    if len(wX) == 0 and len(wR) == 0:
        weights[:] = 1
    elif not wR:
        weights = wX
    elif not wX:
        for t in np.arange(n_trials_X):
            wr = multishift(wR[..., t], shifts, reshape=True).min(axis=1)
            weights[..., t] = wr
    else:
        for t in np.arange(n_trials_X):
            wr = multishift(wR[..., t], shifts, reshape=True).min(axis=1)
            wr = np.amin((wr, wX[:wr.shape[0], :, t]), axis=0)
            weights[..., t] = wr

    wX = weights
    wR = weights

    # remove weighted means
    X, mean1 = demean(X, wX, return_mean=True, inplace=True)
    R = demean(R, wR, inplace=True)

    # equalize power of R channels, the equalize power of the R PCs
    # if R.shape[1] > 1:
    R = normcol(R, wR)
    C, _ = tscov(R)
    V, _ = pca(C, thresh=1e-6)
    z = np.zeros((n_samples_X, V.shape[1], n_trials_R))
    for t in range(n_trials_R):
        z[..., t] = R[..., t] @ V
    R = normcol(z, wR)

    # covariances and cross-covariance with time-shifted refs
    Cr, twcr = tscov(R, shifts, wR)
    Cxr, twcxr = tsxcov(X, R, shifts, wX)

    # regression matrix of x on time-shifted refs
    regression = regcov(Cxr / twcxr, Cr / twcr, keep, thresh)

    # TSPCA: clean x by removing regression on time-shifted refs
    y = np.zeros((n_samples_X, n_chans_X, n_trials_X))
    for t in np.arange(n_trials_X):
        z = multishift(R[..., t], shifts, reshape=True) @ regression
        y[..., t] = X[:z.shape[0], :, t] - z

    y, mean2 = demean(y, wX, return_mean=True, inplace=True)
    idx = np.arange(offset1, initial_samples - offset2)
    mean_total = mean1 + mean2
    weights = wR

    if ndims < 3:
        y = y.squeeze(2)
        weights = weights.squeeze(2)

    return y, idx, mean_total, weights



"""Canonical Correlation Analysis."""
import numpy as np
from scipy import linalg

from .utils import cov_lags, pca
from .utils.matrix import _check_shifts, _times_to_delays, normcol, relshift

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(*args, **kwargs):  # noqa
        if args:
            return args[0]
        return kwargs.get("iterable", None)


def mcca(C, n_channels, n_keep=[]):
    """Multiway canonical correlation analysis.

    As described in [1]_.

    Parameters
    ----------
    C : array, shape=(n_channels * n_datasets, n_channels * n_datasets)
        Covariance matrix of aggregated data sets.
    n_channels : int
        Number of channels of each data set.
    n_keep: int
        Number of components to keep (for orthogonal transforms).

    Returns
    -------
    A : array, shape=(n_channels * n_datasets, n_channels * n_datasets)
        Transform matrix.
    scores : array, shape=(n_comps,)
        Commonality score (ranges from 1 to N^2).
    AA : list of arrays, shapes = (n_channels, n_channels * n_datasets)
        Subject-specific MCCA transform matrices.

    References
    ----------
    .. [1] de Cheveigne, A., Di Liberto, G. M., Arzounian, D., Wong, D.,
       Hjortkjaer, J., Fuglsang, S. A., & Parra, L. C. (2018). Multiway
       Canonical Correlation Analysis of Brain Signals. bioRxiv, 344960.

    """
    if C.shape[0] != C.shape[1]:
        raise ValueError("Covariance must be square !")
    if np.mod(C.shape[0], n_channels) != 0:
        raise ValueError("!")

    # Whiten covariance by blocks
    n_blocks = C.shape[0] // n_channels
    A = np.zeros((n_channels * n_blocks, n_channels * n_blocks))
    for b in range(n_blocks):

        # Extract block covariance
        ix0 = b * n_channels
        ix1 = ix0 + n_channels
        CC = C[ix0:ix1, ix0:ix1]

        # Sphere it
        W = whiten_nt(CC, keep=True)
        A[ix0:ix1, ix0:ix1] = W

    C = A.T @ C @ A

    # final PCA
    V, d = pca(C, thresh=None)  # don't threshold the PCA to keep n_channels
    A = A @ V
    C = V.T @ C @ V
    scores = np.diag(C)

    AA = []
    for b in range(n_blocks):
        AA.append(A[n_channels * b + np.arange(n_channels), :])

    return A, scores, AA


def cca_crossvalidate(xx, yy, shifts=None, sfreq=1, surrogate=False, plot=False):
    """CCA with cross-validation.

    Parameters
    ----------
    xx : list of arrays
        If a list is provided, each element should have shape=(n_times,
        n_chans). If array, it should be 3D of shape=(n_times, n_chans,
        n_trials).
    yy : list of arrays
        If a list is provided, each element should have shape=(n_times,
        n_chans). If array, it should be 3D of shape=(n_times, n_chans,
        n_trials).
    shifts : array, shape=(n_shifts,)
        Array of shifts to apply to `y` relative to `x` (can be negative).
    sfreq : float
        Sampling frequency. If not 1, lags are assumed to be given in seconds.
    surrogate : bool
        If True, estimate SD of correlation over non-matching pairs.
    plot : bool
        Produce some plots.

    Returns
    -------
    AA, BB : arrays
        Cell arrays of transform matrices.
    RR : array, shape=(n_comps, n_shifts, n_trials)
        Correlations (2D).
    SD : array
        Standard deviation of correlation over non-matching pairs (2D).

    """
    if isinstance(xx, list) and isinstance(yy, list):
        assert len(xx) == len(yy)
    elif isinstance(xx, np.ndarray) and isinstance(yy, np.ndarray):
        assert xx.shape[-1] == yy.shape[-1]
        # Convert xx and yy to lists
        xx = [xx[..., t] for t in np.arange(xx.shape[-1])]
        yy = [yy[..., t] for t in np.arange(yy.shape[-1])]
    else:
        raise AttributeError("xx and yy both must be lists of same length, "
                             "or arrays os same n_trials.")

    shifts = _times_to_delays(shifts, sfreq)
    shifts, n_shifts = _check_shifts(shifts)
    n_trials = len(xx)
    n_feats = xx[0].shape[1] + yy[0].shape[1]  # sum of channels

    # Calculate covariance matrices
    print("Calculate all covariances...")
    C = np.zeros((n_feats, n_feats, n_shifts, n_trials)).squeeze()
    for t in tqdm(np.arange(n_trials)):
        C[..., t], _, _ = cov_lags(xx[t], yy[t], shifts)

    # Calculate leave-one-out CCAs
    print("Calculate CCAs...")
    AA = []
    BB = []
    for t in tqdm(np.arange(n_trials)):
        # covariance of all trials except t
        CC = np.sum(C[..., np.arange(n_trials) != t], axis=-1, keepdims=True)
        if CC.ndim == 4:
            CC = np.squeeze(CC, 3)

        # corresponding CCA
        A, B, _ = nt_cca(None, None, None, CC, xx[0].shape[1])
        AA.append(A)
        BB.append(B)
        del A, B

    del C, CC

    # Calculate leave-one-out correlation coefficients
    print("Calculate cross-correlations...")
    n_comps = AA[0].shape[1]
    r = np.zeros((n_comps, n_shifts))
    RR = np.zeros((n_comps, n_shifts, n_trials))
    for t in tqdm(np.arange(n_trials)):
        for s in np.arange(n_shifts):
            x, y = relshift(xx[t], yy[t], shifts[s])
            a = AA[t][:, :, s]
            b = BB[t][:, :, s]
            r[:, s] = np.diag(np.dot(normcol(np.dot(x, a)).T,
                                     normcol(np.dot(y, b)))) / x.shape[0]
            # tt = np.dot(x, a).T
            # tu = np.dot(y, b).T
            # for i in range(tt.shape[0]):
            #     r[i, s] = np.corrcoef(tt[i, :], tu[i, :])[0, 1]

        RR[:, :, t] = r

        # if surrogate:
        #     ss = np.zeros_like(RR)
        #     next = np.mod(t, n_trials)  # correlate with next in list
        #     for s in np.arange(n_shifts):
        #         [x, y] = relshift(xx[t], yy[next], shifts[s])
        #         a = A[:, :, s]
        #         b = B[:, :, s]
        #         m = np.min((x.shape[0], y.shape[0]))
        #         s[:, s] = np.diag(np.dot(normcol(np.dot(x, a)).T,
        #                                  normcol(np.dot(y, b)))) / m
        #     ss[:, :, t] = s

    # if surrogate:
    #     var = (np.sum(ss ** 2, 2) - np.sum(ss, 2) ** 2 / n_trials) /\
    #           (n_trials - 1)
    #     SD = np.sqrt(var)

    if plot:
        import matplotlib.pyplot as plt
        f, (ax1) = plt.subplots(1, 1)
        for k in range(RR.shape[0]):
            ax1.plot(shifts, np.mean(RR[k, :, :], 1).T, label=f"CC{k}")
        ax1.set_title("correlation for each CC")
        ax1.set_xlabel("shift")
        ax1.set_ylabel("correlation")
        ax1.legend()
        # if surrogate:
        #     ax1.plot(SD.T, ':')

        f2, axes = plt.subplots(min(4, RR.shape[0]), 1)
        for k, ax in zip(np.arange(min(4, RR.shape[0])), axes, strict=True):
            idx = np.argmax(np.mean(RR[k, :, :], 1))
            [x, y] = relshift(xx[0], yy[0], shifts[idx])
            ax.plot(np.dot(x, AA[0][:, k, idx]).T, label=f"CC{k}")
            ax.plot(np.dot(y, BB[0][:, k, idx]).T, ":")
            ax.legend()

        ax.set_xlabel("sample")
        f2.set_tight_layout(True)
        plt.show()

    return AA, BB, RR


def nt_cca(X=None, Y=None, lags=None, C=None, m=None, thresh=1e-12, sfreq=1):
    """Compute CCA from covariance.

    Parameters
    ----------
    X, Y : arrays, shape=(n_times, n_chans[, n_trials])
        Data.
    lags : array, shape=(n_lags,)
        Array of lags. A positive lag means Y delayed relative to X. If
        :attr:`sfreq` is > 1, lags are interpreted as times in seconds.
    C : array, shape=(n_chans, n_chans[, n_lags])
        Covariance matrix of [X, Y]. C can be 3D, which case CCA is derived
        independently from each page.
    m : int
        Number of channels of X.
    thresh : float
        Discard principal components below this value.
    sfreq : float
        Sampling frequency. If not 1, lags are assumed to be given in seconds.

    Returns
    -------
    A : array, shape=(n_chans_X, min(n_chans_X, n_chans_Y)[, n_lags])
        Transform matrix mapping `X` to canonical space, where `n_comps` is
        equal to `min(n_chans_X, n_chans_Y)`.
    B : array,  shape=(n_chans_Y, n_comps[, n_lags])
        Transform matrix mapping `Y` to canonical space, where `n_comps` is
        equal to `min(n_chans_X, n_chans_Y)`.
    R : array, shape=(n_comps, n_lags)
        Correlation scores.

    Notes
    -----
    Usage 1: CCA of X, Y
    >> A, B, R = nt_cca(X, Y)  # noqa

    Usage 2: CCA of X, Y for each value of lags.
    >> A, B, R = nt_cca(X, Y, lags)  # noqa

    A positive lag indicates that Y is delayed relative to X.

    Usage 3: CCA from covariance matrix
    >> C = [X, Y].T * [X, Y]  # noqa
    >> A, B, R = nt_cca(None, None, None, C=C, m=X.shape[1])  # noqa

    Use the third form to handle multiple files or large data (covariance C can
    be calculated chunk-by-chunk).

    .. warning:: Means of X and Y are NOT removed.
    .. warning:: A, B are scaled so that (X * A)^2 and (Y * B)^2 are identity
                 matrices (differs from sklearn).

    """
    if (X is None and Y is not None) or (Y is None and X is not None):
        raise AttributeError("Either *both* X and Y should be defined, or C!")

    if X is not None:
        lags = _times_to_delays(lags, sfreq)
        lags, n_lags = _check_shifts(lags)
        C, _, m = cov_lags(X, Y, lags)
        A, B, R = nt_cca(None, None, None, C, m, thresh)
        return A, B, R

    if C is None:
        raise RuntimeError("covariance matrix should be defined")
    if m is None:
        raise RuntimeError("m should be defined")
    if C.shape[0] != C.shape[1]:
        raise RuntimeError("covariance matrix should be square")
    if any((X, Y, lags)):
        raise RuntimeError("only covariance should be defined at this point")
    if C.ndim > 3:
        raise RuntimeError("covariance should be 3D at most")

    if C.ndim == 3:  # covariance is 3D: do a separate CCA for each page
        n_chans, _, n_lags = C.shape
        n_comps = np.min((m, n_chans - m))
        A = np.zeros((m, n_comps, n_lags))
        B = np.zeros((n_chans - m, n_comps, n_lags))
        R = np.zeros((n_comps, n_lags))

        for k in np.arange(n_lags):
            AA, BB, RR = nt_cca(None, None, None, C[:, :, k], m, thresh)
            A[:AA.shape[0], :AA.shape[1], k] = AA
            B[:BB.shape[0], :BB.shape[1], k] = BB
            R[:RR.size, k] = RR
            del AA, BB, RR

        return A, B, R

    # Calculate CCA given C = [X,Y].T * [X,Y] and m = x.shape[1]
    # -------------------------------------------------------------------------
    Wx = whiten_nt(C[:m, :m], thresh)  # sphere X
    Wy = whiten_nt(C[m:, m:], thresh)  # sphere Y

    # apply sphering matrices to C
    W = np.zeros((Wx.shape[0] + Wy.shape[0], Wx.shape[1] + Wy.shape[1]))
    W[:Wx.shape[0], :Wx.shape[1]] = Wx
    W[Wx.shape[0]:, Wx.shape[1]:] = Wy
    C = np.dot(np.dot(W.T, C), W)

    # Number of CCA components
    N = np.min((Wx.shape[1], Wy.shape[1]))

    # PCA
    V, d = pca(C)

    A = np.dot(Wx, V[:Wx.shape[1], :N]) * np.sqrt(2)
    B = np.dot(Wy, V[Wx.shape[1]:, :N]) * np.sqrt(2)
    R = d[:N] - 1

    return A, B, R


def whiten(C, fudge=1e-18):
    """Whiten covariance matrix C of X.

    If X has shape=(observations, components), X_white = np.dot(X, W).

    References
    ----------
    https://stackoverflow.com/questions/6574782/how-to-whiten-matrix-in-pca

    """
    eigvals, V = linalg.eigh(C)  # eigenvalue decomposition of the covariance

    # a fudge factor can be used so that eigenvectors associated with
    # small eigenvalues do not get overamplified.
    D = np.diag(1. / np.sqrt(eigvals + fudge))
    W = np.dot(np.dot(V, D), V.T)   # whitening matrix

    return W


def whiten_nt(C, thresh=1e-12, keep=False):
    """Covariance whitening function from noisetools.

    Parameters
    ----------
    C : array
        Covariance matrix.
    thresh : float
        PCA threshold.
    keep : bool
        If True, infrathreshold components are set to zero. If False (default),
        infrathreshold components are truncated.

    """
    d, V = linalg.eigh(C)  # eigh if matrix symmetric, eig otherwise
    d = np.real(d)
    V = np.real(V)

    # Sort eigenvalues
    idx = np.argsort(d)[::-1]
    d = d[idx]
    V = V[:, idx]

    # Remove small eigenvalues
    good = (d / np.max(d)) > thresh
    if keep is True:
        d[~good] = 0
        V[:, ~good] = 0
    else:
        d = d[good]
        V = V[:, good]

    # break symmetry when x and y perfectly correlated (otherwise cols of x*A
    # and y*B are not orthogonal)
    d = d ** (1 - thresh)
    d_norm = d / np.max(d)

    dd = np.zeros_like(d)
    dd[d_norm > thresh] = (1. / d[d_norm > thresh])

    D = np.diag(np.sqrt(dd))
    W = np.dot(V, D)

    return W


def whiten_svd(X):
    """SVD whitening."""
    U, S, Vt = linalg.svd(X, full_matrices=False)

    # U and Vt are the singular matrices, and s contains the singular values.
    # Since the rows of both U and Vt are orthonormal vectors, then U * Vt
    # will be white
    X_white = np.dot(U, Vt)

    return X_white


def whiten_zca(C, thresh=None):
    """Compute ZCA whitening matrix (aka Mahalanobis whitening).

    Parameters
    ----------
    C : array
        Covariance matrix.
    thresh : float
        Whitening constant, it prevents division by zero.

    Returns
    -------
    ZCA: array, shape (n_chans, n_chans)
        ZCA matrix, to be multiplied with data.

    """
    U, S, V = np.linalg.svd(C)  # Singular Value Decomposition

    # ZCA Whitening matrix
    D = np.diag(1. / np.sqrt(S + thresh))
    ZCA = np.dot(np.dot(U, D), U.T)

    return ZCA



import matplotlib.pyplot as plt
import numpy as np

from meegkit import dss, sns, tspca
from meegkit.utils import demean, fold, unfold

rng = np.random.default_rng(9)


def test_tspca_sns_dss():  # TODO
    """Test TSPCA, SNS, DSS.

    Requires data stored in a time X channels X trials matrix.

    Remove environmental noise with TSPCA (shifts=-50:50).
    Remove sensor noise with SNS.
    Remove non-repeatable components with DSS.
    """
    # Random data (time*chans*trials)
    data = rng.random((800, 102, 200))
    ref = rng.random((800, 3, 200))

    # remove means
    noisy_data = demean(data)

    # Apply TSPCA
    # -------------------------------------------------------------------------
    shifts = np.arange(-50, 51)
    print("TSPCA...")
    y_tspca, idx = tspca.tsr(noisy_data, ref, shifts)[0:2]
    print("\b OK!")
    y_tspca = noisy_data

    # Apply SNS
    # -------------------------------------------------------------------------
    nneighbors = 10
    print("SNS...")
    y_tspca_sns, r = sns.sns(y_tspca, nneighbors)
    print("\b OK!")

    # apply DSS
    # -------------------------------------------------------------------------
    print("DSS...")
    # Keep all PC components
    y_tspca_sns = demean(y_tspca_sns)
    print(y_tspca_sns.shape)
    todss, fromdss, _, _ = dss.dss1(y_tspca_sns)
    print("\b OK!")

    # c3 = DSS components
    y_tspca_sns_dss = fold(
        np.dot(unfold(y_tspca_sns), todss), y_tspca_sns.shape[0])

    # TODO do something with it
    assert y_tspca_sns_dss.shape == noisy_data.shape


def test_tsr(show=True):
    """Test time-shift regression."""
    sr = 200
    nsamples = 10000
    nchans = 10
    x = rng.standard_normal((nsamples, nchans))

    # artifact + harmonics
    artifact = np.sin(np.arange(nsamples) / sr * 2 * np.pi * 10)[:, None]
    artifact[artifact < 0] = 0
    artifact = 5 * artifact ** 3
    signal = x + artifact

    signal -= np.mean(signal, keepdims=True)
    artifact -= np.mean(artifact, keepdims=True)

    # Without shifts
    y, idx, mean_total, weights = tspca.tsr(
        signal,
        artifact,
        shifts=[0])

    if show:
        f, ax = plt.subplots(2, 1, num="without shifts")
        ax[0].plot(y[:500, 0], "grey", label="recovered signal")
        ax[0].plot(x[:500, 0], ":", label="real signal")
        ax[1].plot((y - x)[:500], label="residual")
        ax[0].legend()
        ax[1].set_xlabel("time (samples)")
        ax[0].set_title("signals")
        ax[1].set_title("residuals")
        plt.tight_layout()
        # ax[1].legend()
        # plt.show()

    # Test residual almost 0.0
    np.testing.assert_almost_equal(y - x, np.zeros_like(y), decimal=1)

    # With shifts. We slide the input array by one sample, and check that the
    # artifact is successfully regressed.
    y, idx, mean_total, weights = tspca.tsr(
        signal,
        np.roll(artifact, 1, axis=0),
        shifts=[-1, 0, 1])

    if show:
        f, ax = plt.subplots(3, 1, num="with shifts")
        ax[0].plot(signal[:500], label="signal + noise")
        ax[1].plot(x[:500, 0], "grey", label="real signal")
        ax[1].plot(y[:500, 0], ":", label="recovered signal")
        ax[2].plot((signal - y)[:500, 0], label="before - after")
        # ax[0].legend()
        ax[1].legend()
        ax[2].legend()
        ax[1].set_xlabel("time (samples)")
        plt.tight_layout()
        plt.show()

if __name__ == "__main__":
    import pytest
    pytest.main([__file__])
    # test_tspca_sns_dss()
    # test_tsr()



"""Test RESS."""
import matplotlib.pyplot as plt
import numpy as np
import pytest
import scipy.signal as ss
from scipy.linalg import pinv

from meegkit import ress
from meegkit.utils import fold, matmul3d, snr_spectrum, unfold

rng = np.random.default_rng(9)


def create_data(n_times, n_chans=10, n_trials=20, freq=12, sfreq=250,
                noise_dim=8, SNR=.8, t0=100, show=False):
    """Create synthetic data.
    Returns
    -------
    noisy_data: array, shape=(n_times, n_channels, n_trials)
        Simulated data with oscillatory component strting at t0.
    """
    # source
    source = np.sin(2 * np.pi * freq * np.arange(n_times - t0) / sfreq)[None].T
    s = source * rng.standard_normal((1, n_chans))
    s = s[:, :, np.newaxis]
    s = np.tile(s, (1, 1, n_trials))
    signal = np.zeros((n_times, n_chans, n_trials))
    signal[t0:, :, :] = s

    # noise
    noise = np.dot(
        unfold(rng.standard_normal((n_times, noise_dim, n_trials))),
        rng.standard_normal((noise_dim, n_chans)))
    noise = fold(noise, n_times)

    # mix signal and noise
    signal = SNR * signal /  np.sqrt(np.mean(signal ** 2))
    noise = noise / np.sqrt(np.mean(noise ** 2))
    noisy_data = signal + noise

    if show:
        f, ax = plt.subplots(3)
        ax[0].plot(signal[:, 0, 0], label="source")
        ax[1].plot(noise[:, 1, 0], label="noise")
        ax[2].plot(noisy_data[:, 1, 0], label="mixture")
        ax[0].legend()
        ax[1].legend()
        ax[2].legend()
        plt.show()

    return noisy_data, signal


@pytest.mark.parametrize("target", [12, 15, 20])
@pytest.mark.parametrize("n_trials", [16])
@pytest.mark.parametrize("peak_width", [.5, 1])
@pytest.mark.parametrize("neig_width", [1])
@pytest.mark.parametrize("neig_freq", [1])
def test_ress(target, n_trials, peak_width, neig_width, neig_freq, show=False):
    """Test RESS."""
    sfreq = 250
    n_keep = 1
    n_chans = 10
    n_times = 1000
    data, source = create_data(n_times=n_times, n_trials=n_trials,
                               n_chans=n_chans, freq=target, sfreq=sfreq,
                               show=False)
    r = ress.RESS(sfreq=sfreq, peak_freq=target, neig_freq=neig_freq,
        peak_width=peak_width, neig_width=neig_width, n_keep=n_keep,
        compute_unmixing=True)
    out = r.fit_transform(data)

    nfft = 500
    bins, psd = ss.welch(np.squeeze(out), sfreq, window="boxcar",
                         nperseg=nfft / (peak_width * 2),
                         noverlap=0, axis=0, average="mean")
    # psd = np.abs(np.fft.fft(out, nfft, axis=0))
    # psd = psd[0:psd.shape[0] // 2 + 1]
    # bins = np.linspace(0, sfreq // 2, psd.shape[0])
    # print(psd.shape)
    # print(bins[:10])

    psd = psd.mean(axis=-1, keepdims=True)  # average over trials
    snr = snr_spectrum(psd + psd.max() / 20, bins, skipbins=1, n_avg=2)
    # snr = snr.mean(1)
    if show:
        f, ax = plt.subplots(2)
        ax[0].plot(bins, snr, ":o")
        ax[0].axhline(1, ls=":", c="grey", zorder=0)
        ax[0].axvline(target, ls=":", c="grey", zorder=0)
        ax[0].set_ylabel("SNR (a.u.)")
        ax[0].set_xlabel("Frequency (Hz)")
        ax[0].set_xlim([0, 40])
        ax[0].set_ylim([0, 10])
        ax[1].plot(bins, psd)
        ax[1].axvline(target, ls=":", c="grey", zorder=0)
        ax[1].set_ylabel("PSD")
        ax[1].set_xlabel("Frequency (Hz)")
        ax[1].set_xlim([0, 40])
        # plt.show()

    assert snr[bins == target] > 10
    assert (snr[(bins <= target - 2) | (bins >= target + 2)] < 2).all()

    # test multiple components

    out = r.transform(data)
    toress = r.to_ress
    fromress = r.from_ress

    proj = matmul3d(out, fromress)
    assert proj.shape == (n_times, n_chans, n_trials)

    if show:
        f, ax = plt.subplots(data.shape[1], 2, sharey="col")
        for c in range(data.shape[1]):
            ax[c, 0].plot(data[:, c].mean(-1), lw=.5, label="data")
            ax[c, 1].plot(proj[:, c].mean(-1), lw=.5, label="projection")
            if c < data.shape[1]:
                ax[c, 0].set_xticks([])
                ax[c, 1].set_xticks([])

        ax[0, 0].set_title("Before")
        ax[0, 1].set_title("After")
        plt.legend()

    # 2 comps
    _ = ress.RESS(
        sfreq=sfreq, peak_freq=target, n_keep=2
    ).fit_transform(data)

    # All comps
    r = ress.RESS(sfreq=sfreq, peak_freq=target, neig_freq=neig_freq,
        peak_width=peak_width, neig_width=neig_width, n_keep=-1, compute_unmixing=True)
    out = r.fit_transform(data)
    toress = r.to_ress
    fromress = r.from_ress

    if show:
        # Inspect mixing/unmixing matrices
        combined_data = np.array([toress, fromress, pinv(toress)])
        _max = np.amax(combined_data)

        f, ax = plt.subplots(3)
        ax[0].imshow(toress, label="toRESS")
        ax[0].set_title("toRESS")
        ax[1].imshow(fromress, label="fromRESS", vmin=-_max, vmax=_max)
        ax[1].set_title("fromRESS")
        ax[2].imshow(pinv(toress), vmin=-_max, vmax=_max)
        ax[2].set_title("toRESS$^{-1}$")
        plt.tight_layout()
        plt.show()

    print(np.sum(np.abs(pinv(toress) - fromress) >= .1))


if __name__ == "__main__":
    import pytest
    # pytest.main([__file__])
    test_ress(20, 16, 1, 1, 1, show=True)



"""Math utils."""
from scipy import linalg


def mrdivide(A, B):
    r"""Matrix right-division (A/B).

    Solves the linear system XB = A for X. We can write equivalently:

    1) XB = A
    2) (XB).T = A.T
    3) B.T X.T = A.T

    Therefore A/B amounts to solving B.T X.T = A.T for X.T:

    >> mldivide(B.T, A.T).T

    References
    ----------
    https://docs.scipy.org/doc/numpy/user/numpy-for-matlab-users.html

    """
    return mldivide(B.T, A.T).T


def mldivide(A, B):
    r"""Matrix left-division (A\B).

    Solves the AX = B for X. In other words, X minimizes norm(A*X - B), the
    length of the vector AX - B:
    - linalg.solve(A, B) if A is square
    - linalg.lstsq(A, B) otherwise

    References
    ----------
    https://docs.scipy.org/doc/numpy/user/numpy-for-matlab-users.html

    """
    try:
        # Note: we must use overwrite_a=False in order to be able to
        # use the fall-back solution below in case a LinAlgError is raised
        return linalg.solve(A, B, assume_a="pos", overwrite_a=False)
    except linalg.LinAlgError:
        # Singular matrix in solving dual problem. Using least-squares
        # solution instead.
        try:
            return linalg.lstsq(A, B, lapack_driver="gelsy")[0]
        except linalg.LinAlgError:
            print("Solution not stable. Model not updated!")
            return None



"""M/EEG denoising utilities in python."""
__version__ = "0.1.5"

from . import asr, cca, detrend, dss, lof, ress, sns, star, trca, tspca, utils

__all__ = ["asr", "cca", "detrend", "dss", "lof", "ress", "sns", "star", "trca",
           "tspca", "utils"]



"""Sensor noise suppression."""
import numpy as np

from .tspca import tsr
from .utils import demean, fold, pca, theshapeof, tscov, unfold
from .utils.matrix import _check_weights

DEFAULT_WEIGHTS = np.array([])


def sns(X, n_neighbors=0, skip=0, weights=DEFAULT_WEIGHTS):
    """Sensor Noise Suppression.

    This algorithm will replace the data from each channel by its regression on
    the subspace formed by the other channels. The underlying assumptions are
    that (a) sensor noise is uncorrelated across sensors, and (b) genuine
    signal is correlated, sensor noise is removed and genuine signal is
    preserved.

    Parameters
    ----------
    X : array, shape=(n_times, n_chans, n_trials)
        EEG data.
    n_neighbors : int
        Number of neighbors (based on correlation) to include in the
        projection.
    skip: int
        Number of closest neighbors to skip (default=0).
    weights : array of floats
        Weights (default=all ones).


    Returns
    -------
    y : array, shape=(n_times, n_chans, n_trials)
        Denoised data.
    r : array, shape=(n_chans, n_chans)
        Denoising matrix.

    """
    if not n_neighbors:
        n_neighbors = X.shape[1] - 1
    ndims = X.ndim
    weights = _check_weights(weights, X)
    n_samples, n_chans, n_trials = theshapeof(X)

    X = unfold(X)
    X = demean(X)
    c, nc = tscov(X)

    if weights.any():
        weights = unfold(weights)
        X = demean(X, weights)
        wc, nwc = tscov(X, shifts=None, weights=weights)
        r = sns0(c, n_neighbors, skip, wc)
    else:
        r = sns0(c, n_neighbors, skip, c)

    y = np.dot(np.squeeze(X), r)
    if ndims > 2:
        y = fold(y, n_samples)

    return y, r


def sns0(c, n_neighbors=0, skip=0, wc=DEFAULT_WEIGHTS):
    """Sensor Noise Suppression from data covariance.

    Parameters
    ----------
    c: array, shape=(n_chans, n_chans)
        Full covariance of data to denoise.
    n_neighbors : int
        Number of neighbors (based on correlation) to include in the
        projection.
    skip: int
        Number of closest neighbors to skip (default=0).
    wc: array
        Weighted covariance.

    Returns
    -------
    r: arraym
        Denoising matrix.

    """
    if not wc.any():
        wc = c.copy()

    n_chans = c.shape[0]

    if n_neighbors == 0:
        n_neighbors = n_chans - skip - 1
    else:
        n_neighbors = np.min((n_neighbors, n_chans - skip - 1))

    r = np.zeros(c.shape)

    # normalize
    eps = np.finfo(np.float32).eps
    norm = np.diag(c).copy()
    mask = norm > eps
    norm[mask] = np.sqrt(1. / norm[mask])
    norm[~mask] = 0
    c = c * norm
    c *= norm[:, np.newaxis]
    del norm, mask

    for k in np.arange(n_chans):
        c1 = c[:, k]  # correlation of channel k with all other channels
        # sort by correlation, descending order
        idx = np.argsort(c1 ** 2, 0)[::-1]
        idx = idx[skip + 1:skip + n_neighbors + 1]  # keep best

        # pca neighbors to orthogonalize them
        c2 = wc[idx, :][:, idx]
        [eigvec, eigval] = pca(c2)

        # Some of the eigenvalues could be zero or inf
        norm = np.zeros(len(eigval))
        mask = np.logical_and(eigval > eps, np.isfinite(eigval))
        norm[mask] = 1. / np.sqrt(eigval[mask])
        eigvec *= norm
        del eigval, norm, mask

        r[k, idx] = np.dot(eigvec, np.dot(wc[k][idx], eigvec))

        if r[k, k] != 0:
            raise RuntimeError("SNS operator should be zero along diagonal")

    return r.T


def sns1(X, n_neighbors=None, skip=0):
    """Sensor Noise Suppression 1.

    This version of SNS first regresses out major shared components.
    """
    if X.ndim > 2:
        raise Exception("SNS1 works only with 2D matrices")

    n_samples, n_chans, n_trials = theshapeof(X)

    if n_neighbors is None:
        n_neighbors = n_chans - skip - 1
    else:
        n_neighbors = np.min((n_neighbors, n_chans - skip - 1))

    mn = np.mean(X)
    X = (X - mn)  # remove mean
    N = np.sqrt(np.sum(X ** 2))
    NN = 1 / N
    NN[np.where(np.isnan(NN))] = 0
    X = (X * NN)  # normalize

    y = np.zeros(X.shape)

    for k in np.arange(n_chans):
        c1 = X.T * X[:, k]  # correlation with neighbors
        c1 = c1 / c1[k]
        c1[k] = 0  # demote self
        [c1, idx] = np.sort(c1 ** 2, 0)[::-1]  # sort
        idx = idx[1 + skip:n_neighbors + skip]   # keep best

        # pca neighbors to orthogonalize them
        xx = X[:, idx]
        c2 = xx.T * xx
        [eigvec, eigval] = pca(c2)
        eigvec = eigvec * np.diag(1 / np.sqrt(eigval))

        y[:, k] = tsr(X[:, k], xx * eigvec)

    y = (y * N)

    return y



"""Task-Related Component Analysis."""
# Authors: Giuseppe Ferraro <giuseppe.ferraro@isae-supaero.fr>
#          Ludovic Darmet <ludovic.darmet@isae-supaero.fr>
import numpy as np
import scipy.linalg as linalg
from pyriemann.estimation import Covariances
from pyriemann.utils.mean import mean_covariance

from .utils import theshapeof
from .utils.trca import bandpass, schaefer_strimmer_cov


class TRCA:
    """Task-Related Component Analysis (TRCA).

    Parameters
    ----------
    sfreq : float
        Sampling rate.
    filterbank : list[[2-tuple, 2-tuple]]
        Filterbank frequencies. Each list element is itself a list of passband
        `Wp` and stopband `Ws` edges frequencies `[Wp, Ws]`. For example, this
        creates 3 bands, starting at 6, 14, and 22 hz respectively::

            [[(6, 90), (4, 100)],
             [(14, 90), (10, 100)],
             [(22, 90), (16, 100)]]

        See :func:`scipy.signal.cheb1ord()` for more information on how to
        specify the `Wp` and `Ws`.
    ensemble : bool
        If True, perform the ensemble TRCA analysis (default=False).
    method : str in {'original'| 'riemann'}
        Use original implementation from [1]_ or a variation that uses
        regularization and the geodesic mean [2]_.
    regularization : str in {'schaefer' | 'lwf' | 'oas' | 'scm'}
        Regularization estimator used for covariance estimation with the
        `riemann` method. Consider 'schaefer', 'lwf', 'oas'. 'scm' does not add
        regularization and is almost equivalent to the original implementation.

    Attributes
    ----------
    traindata : array, shape=(n_bands, n_chans, n_trials)
        Reference (training) data decomposed into sub-band components by the
        filter bank analysis.
    y_train : array, shape=(n_trials)
        Labels associated with the train data.
    coef_ : array, shape=(n_chans, n_chans)
        Weight coefficients for electrodes which can be used as a spatial
        filter.
    classes : list
        Classes.
    n_bands : int
        Number of sub-bands.

    References
    ----------
    .. [1] M. Nakanishi, Y. Wang, X. Chen, Y. -T. Wang, X. Gao, and T.-P. Jung,
       "Enhancing detection of SSVEPs for a high-speed brain speller using
       task-related component analysis", IEEE Trans. Biomed. Eng,
       65(1):104-112, 2018.
    .. [2] Barachant, A., Bonnet, S., Congedo, M., & Jutten, C. (2010,
       October). Common spatial pattern revisited by Riemannian geometry. In
       2010 IEEE International Workshop on Multimedia Signal Processing (pp.
       472-476). IEEE.

    """

    def __init__(self, sfreq, filterbank, ensemble=False, method="original",
                 estimator="scm"):
        self.sfreq = sfreq
        self.ensemble = ensemble
        self.filterbank = filterbank
        self.n_bands = len(self.filterbank)
        self.coef_ = None
        self.method = method
        if estimator == "schaefer":
            self.estimator = schaefer_strimmer_cov
        else:
            self.estimator = estimator

    def fit(self, X, y):
        """Training stage of the TRCA-based SSVEP detection.

        Parameters
        ----------
        X : array, shape=(n_samples, n_chans[, n_trials])
            Training EEG data.
        y : array, shape=(trials,)
            True label corresponding to each trial of the data array.

        """
        n_samples, n_chans, _ = theshapeof(X)
        classes = np.unique(y)

        trains = np.zeros((len(classes), self.n_bands, n_samples, n_chans))

        W = np.zeros((self.n_bands, len(classes), n_chans))

        for class_i in classes:
            # Select data with a specific label
            eeg_tmp = X[..., y == class_i]
            for fb_i in range(self.n_bands):
                # Filter the signal with fb_i
                eeg_tmp = bandpass(eeg_tmp, self.sfreq,
                                   Wp=self.filterbank[fb_i][0],
                                   Ws=self.filterbank[fb_i][1])
                if (eeg_tmp.ndim == 3):
                    # Compute mean of the signal across trials
                    trains[class_i, fb_i] = np.mean(eeg_tmp, -1)
                else:
                    trains[class_i, fb_i] = eeg_tmp
                # Find the spatial filter for the corresponding filtered signal
                # and label
                if self.method == "original":
                    w_best = trca(eeg_tmp)
                elif self.method == "riemann":
                    w_best = trca_regul(eeg_tmp, self.estimator)
                else:
                    raise ValueError("Invalid `method` option.")

                W[fb_i, class_i, :] = w_best  # Store the spatial filter

        self.trains = trains
        self.coef_ = W
        self.classes = classes

        return self

    def predict(self, X):
        """Test phase of the TRCA-based SSVEP detection.

        Parameters
        ----------
        X: array, shape=(n_samples, n_chans[, n_trials])
            Test data.

        Returns
        -------
        pred: np.array, shape (trials)
            The target estimated by the method.

        """
        if self.coef_ is None:
            raise RuntimeError("TRCA is not fitted")

        # Alpha coefficients for the fusion of filterbank analysis
        fb_coefs = [(x + 1)**(-1.25) + 0.25 for x in range(self.n_bands)]
        _, _, n_trials = theshapeof(X)

        r = np.zeros((self.n_bands, len(self.classes)))
        pred = np.zeros((n_trials), "int")  # To store predictions

        for trial in range(n_trials):
            test_tmp = X[..., trial]  # pick a trial to be analysed
            for fb_i in range(self.n_bands):

                # filterbank on testdata
                testdata = bandpass(test_tmp, self.sfreq,
                                    Wp=self.filterbank[fb_i][0],
                                    Ws=self.filterbank[fb_i][1])

                for class_i in self.classes:
                    # Retrieve reference signal for class i
                    # (shape: n_chans, n_samples)
                    traindata = np.squeeze(self.trains[class_i, fb_i])
                    if self.ensemble:
                        # shape = (n_chans, n_classes)
                        w = np.squeeze(self.coef_[fb_i]).T
                    else:
                        # shape = (n_chans)
                        w = np.squeeze(self.coef_[fb_i, class_i])

                    # Compute 2D correlation of spatially filtered test data
                    # with ref
                    r_tmp = np.corrcoef((testdata @ w).flatten(),
                                        (traindata @ w).flatten())
                    r[fb_i, class_i] = r_tmp[0, 1]

            rho = np.dot(fb_coefs, r)  # fusion for the filterbank analysis

            tau = np.argmax(rho)  # retrieving index of the max
            pred[trial] = int(tau)

        return pred


def trca(X):
    """Task-related component analysis.

    This function implements the method described in [1]_.

    Parameters
    ----------
    X : array, shape=(n_samples, n_chans[, n_trials])
        Training data.

    Returns
    -------
    W : array, shape=(n_chans,)
        Weight coefficients for electrodes which can be used as a spatial
        filter.

    References
    ----------
    .. [1] M. Nakanishi, Y. Wang, X. Chen, Y. -T. Wang, X. Gao, and T.-P. Jung,
       "Enhancing detection of SSVEPs for a high-speed brain speller using
       task-related component analysis", IEEE Trans. Biomed. Eng,
       65(1):104-112, 2018.

    """
    n_samples, n_chans, n_trials = theshapeof(X)

    # 1. Compute empirical covariance of all data (to be bounded)
    # -------------------------------------------------------------------------
    # Concatenate all the trials to have all the data as a sequence
    UX = np.zeros((n_chans, n_samples * n_trials))
    for trial in range(n_trials):
        UX[:, trial * n_samples:(trial + 1) * n_samples] = X[..., trial].T

    # Mean centering
    UX -= np.mean(UX, 1)[:, None]

    # Covariance
    Q = UX @ UX.T

    # 2. Compute average empirical covariance between all pairs of trials
    # -------------------------------------------------------------------------
    S = np.zeros((n_chans, n_chans))
    for trial_i in range(n_trials - 1):
        x1 = np.squeeze(X[..., trial_i])

        # Mean centering for the selected trial
        x1 -= np.mean(x1, 0)

        # Select a second trial that is different
        for trial_j in range(trial_i + 1, n_trials):
            x2 = np.squeeze(X[..., trial_j])

            # Mean centering for the selected trial
            x2 -= np.mean(x2, 0)

            # Compute empirical covariance between the two selected trials and
            # sum it
            S = S + x1.T @ x2 + x2.T @ x1

    # 3. Compute eigenvalues and vectors
    # -------------------------------------------------------------------------
    lambdas, W = linalg.eig(S, Q, left=True, right=False)

    # Select the eigenvector corresponding to the biggest eigenvalue
    W_best = W[:, np.argmax(lambdas)]

    return W_best


def trca_regul(X, method):
    """Task-related component analysis.

    This function implements a variation of the method described in [1]_. It is
    inspired by a riemannian geometry approach to CSP [2]_. It adds
    regularization to the covariance matrices and uses the riemannian mean for
    the inter-trial covariance matrix `S`.

    Parameters
    ----------
    X : array, shape=(n_samples, n_chans[, n_trials])
        Training data.

    Returns
    -------
    W : array, shape=(n_chans,)
        Weight coefficients for electrodes which can be used as a spatial
        filter.

    References
    ----------
    .. [1] M. Nakanishi, Y. Wang, X. Chen, Y. -T. Wang, X. Gao, and T.-P. Jung,
       "Enhancing detection of SSVEPs for a high-speed brain speller using
       task-related component analysis", IEEE Trans. Biomed. Eng,
       65(1):104-112, 2018.
    .. [2] Barachant, A., Bonnet, S., Congedo, M., & Jutten, C. (2010,
       October). Common spatial pattern revisited by Riemannian geometry. In
       2010 IEEE International Workshop on Multimedia Signal Processing (pp.
       472-476). IEEE.

    """
    n_samples, n_chans, n_trials = theshapeof(X)

    # 1. Compute empirical covariance of all data (to be bounded)
    # -------------------------------------------------------------------------
    # Concatenate all the trials to have all the data as a sequence
    UX = np.zeros((n_chans, n_samples * n_trials))
    for trial in range(n_trials):
        UX[:, trial * n_samples:(trial + 1) * n_samples] = X[..., trial].T

    # Mean centering
    UX -= np.mean(UX, 1)[:, None]

    # Compute empirical variance of all data (to be bounded)
    cov = Covariances(estimator=method).fit_transform(UX[np.newaxis, ...])
    Q = np.squeeze(cov)

    # 2. Compute average empirical covariance between all pairs of trials
    # -------------------------------------------------------------------------
    # Intertrial correlation computation
    data = np.concatenate((X, X), axis=1)

    # Swapaxes to fit pyriemann Covariances
    data = np.swapaxes(data, 0, 2)
    cov = Covariances(estimator=method).fit_transform(data)

    # Keep only inter-trial
    S = cov[:, :n_chans, n_chans:] + cov[:, n_chans:, :n_chans]

    # If the number of samples is too big, we compute an approximate of
    # riemannian mean to speed up the computation
    if n_trials < 30:
        S = mean_covariance(S, metric="riemann")
    else:
        S = mean_covariance(S, metric="logeuclid")

    # 3. Compute eigenvalues and vectors
    # -------------------------------------------------------------------------
    lambdas, W = linalg.eig(S, Q, left=True, right=False)

    # Select the eigenvector corresponding to the biggest eigenvalue
    W_best = W[:, np.argmax(lambdas)]

    return W_best



"""Robust detrending."""
import numpy as np
from scipy.signal import lfilter

from .utils import demean, mrdivide, pca, unfold
from .utils.matrix import _check_weights
from .utils.sig import stmcb


def detrend(x, order, w=None, basis="polynomials", threshold=3, n_iter=4,
            show=False):
    """Robustly remove trend.

    The data are fit to the basis using weighted least squares. The weight is
    updated by setting samples for which the residual is greater than 'thresh'
    times its std to zero, and the fit is repeated at most 'niter'-1 times.

    The choice of order (and basis) determines what complexity of the trend
    that can be removed.  It may be useful to first detrend with a low order
    to avoid fitting outliers, and then increase the order.

    Parameters
    ----------
    x : array, shape=(n_times, n_channels[, n_trials])
        Raw data matrix.
    order : int
        Order of polynomial or number of sin/cosine pairs
    w : weights, shape=(n_times[, n_channels][, n_trials])
        Sample weights for the regression. If a single channel is provided, the
        same weights are applied to all channels.
    basis : {'polynomials', 'sinusoids'} | ndarray
        Basis for regression.
    threshold : int
        Threshold for outliers, in number of standard deviations (default=3).
    niter : int
        Number of iterations (default=5).

    Returns
    -------
    y : array, shape=(n_times, n_channels[, n_trials])
        Detrended data.
    w : array, shape=(n_times[, n_channels][, n_trials])
        Updated weights.
    r : array, shape=(n_times * ntrials, order)
        Basis matrix used.

    Examples
    --------
    Fit a linear trend, ignoring samples > 3*sd from it, and remove it:
    >> y = detrend(x, 1)

    Fit/remove polynomial (order=5) with initial weighting w, threshold = 4*sd:
    >> y = detrend(x, 5, w, 'polynomial', 4)

    Fit/remove linear then 3rd order polynomial trend:
    >> [y, w]= detrend(x, 1)
    >> [yy, ww] = detrend(y, 3)

    """
    if threshold == 0:
        raise ValueError("thresh=0 is not what you want...")

    # check/fix sizes
    dims = x.shape
    w = _check_weights(w, x)
    x = unfold(x)
    w = unfold(w)
    n_times, n_chans = x.shape

    # regressors
    if isinstance(basis, np.ndarray):
        r = basis
    else:
        lin = np.linspace(-1, 1, n_times)
        if basis == "polynomials" or basis is None:
            r = np.zeros((n_times, order))
            for i, o in enumerate(range(1, order + 1)):
                r[:, i] = lin ** o
        elif basis == "sinusoids":
            r = np.zeros((n_times, order * 2))
            for i, o in enumerate(range(1, order + 1)):
                r[:, 2 * i] = np.sin(2 * np.pi * o * lin / 2)
                r[:, 2 * i + 1] = np.cos(2 * np.pi * o * lin / 2)
        else:
            raise ValueError("!")

    # iteratively remove trends
    # the tricky bit is to ensure that weighted means are removed before
    # calculating the regression (see regress()).
    for _ in range(n_iter):
        # weighted regression on basis
        _, y = regress(x, r, w)

        # find outliers
        d = x - y
        if w.any():
            d = d * w
        ww = np.ones_like(x)
        ww[(abs(d) > threshold * np.std(d))] = 0

        # update weights
        if not w.any():
            w = ww
        else:
            w = np.amin((w, ww), axis=0)
        del ww

    y = x - y
    y = np.reshape(y, dims)
    w = np.reshape(w, dims)

    if show:
        _plot_detrend(x, y, w)

    return y, w, r


def regress(x, r, w=None, threshold=1e-7, return_mean=False):
    """Weighted regression.

    Parameters
    ----------
    x : array, shape=(n_times, n_chans)
        Data.
    r : array, shape=(n_times, n_chans)
        Regressor.
    w : array, shape=(n_times, n_chans)
        Weight to apply to `x`. `w` is either a matrix of same size as `x`, or
        a column vector to be applied to each column of `x`.
    threshold : float
        PCA threshold (default=1e-7). Dimensions of x with eigenvalue lower
        than this value will be discarded.
    return_mean : bool
        If True, also return the signal mean prior to regression.

    Returns
    -------
    b : array, shape=(n_chans, n_chans)
        Regression matrix (apply to r to approximate x).
    z : array, shape=(n_times, n_chans)
        Regression (r @ b).

    """
    # check/fix sizes
    w = _check_weights(w, x)
    n_times = x.shape[0]
    n_chans = x.shape[1]
    n_regs = r.shape[1]
    r = unfold(r)
    x = unfold(x)
    if r.shape[0] != x.shape[0]:
        raise ValueError("r and x have incompatible shapes!")

    # save weighted mean
    mn = x - demean(x, w)

    if not w.any():  # simple regression
        rr = demean(r)
        yy = demean(x)

        # PCA
        V, _ = pca(rr.T @ rr, thresh=threshold)
        rrr = rr.dot(V)

        # Regression (OLS)
        b = mrdivide(yy.T.dot(rrr), rrr.T.dot(rrr))
        b = b.T
        z = np.dot(demean(r, w).dot(V), b)
        z = z + mn

    else:  # weighted regression
        if w.shape[0] != n_times:
            raise ValueError("!")

        if w.shape[1] == 1:  # same weight for all channels
            if sum(w.flatten()) == 0:
                print("weights all zero")
                b = 0
            else:
                yy = demean(x, w) * w
                rr = demean(r, w) * w
                V, _ = pca(rr.T @ rr, thresh=threshold)
                rr = rr @ V
                b = mrdivide(yy.T @ rr, rr.T @ rr)

            z = demean(r, w).dot(V).dot(b.T)
            z = z + mn

        else:  # each channel has own weight
            if w.shape[1] != x.shape[1]:
                raise ValueError("!")
            z = np.zeros(x.shape)
            b = np.zeros((n_chans, n_regs))
            for i in range(n_chans):
                if not np.any(w[:, i]):
                    print(f"weights are all zero for channel {i}")
                else:
                    wc = w[:, i][:, None]  # channel-specific weight
                    xx = demean(x[:, i], wc) * wc

                    # remove channel-specific-weighted mean from regressor
                    r = demean(r, wc, inplace=True)
                    rr = r * wc
                    V, _ = pca(rr.T @ rr, thresh=threshold)
                    rr = rr.dot(V)
                    b[i, :V.shape[1]] = mrdivide(xx.T @ rr, rr.T @ rr)

                z[:, i] = np.squeeze(r @ (V @ b[i, :V.shape[1]].T)) + mn[:, i]

    return b, z


def reduce_ringing(X, samples, order=10, n_samples=100, extra=50, threshold=3,
                   show=False):
    """Subtract filter impulse response from signal at given samples.

    Parameters
    ----------
    X: ndarray, shape=(n_times, n_chans[, n_trials])
        Data containing ringing artifacts.
    samples : list of ints
        Sample indices where to find ringing artifacts.
    order : int
        Order of polynomial trend (default=10).
    n_samples = 100
        Number of samples over which to estimate impulse response
        (default=100).
    extra : int
        Samples before stimulus to anchor trend (default=50).
    threshold: float
        Threshold for robust detrending (default=3).

    Returns
    -------
    y : ndarray, shape=(n_times, n_chans[, n_trials])
        Clean data.

    """
    NNUM = 8
    NDEN = 8  # number of filter coeffs

    # remove samples too close to beginning or end
    samples = samples[samples > extra]
    samples = samples[samples < X.shape[0] - n_samples]

    y = X.copy()
    for _i, s in enumerate(samples):
        for c in range(X.shape[1]):
            # select portion to fit filter response, remove polynomial trend
            response = X[s - extra:s + n_samples, c]
            # response = detrend(response, order, threshold)
            response = response[extra:]

            # estimate filter parameters - helps ensure stable filter
            response = np.r_[(response, np.zeros(response.shape))]
            [B, A] = stmcb(response, q=NNUM, p=NDEN, niter=20)

            # estimate filter response to event
            pulse = np.arange(n_samples) < 1
            model = lfilter(B, A, pulse)
            idx = s + np.arange(model.shape[0])
            y[idx, c] = X[idx, c] - model

    if show:
        w = np.zeros((X.shape[0], X.shape[1]))
        for s in samples:
            w[s:s + n_samples, :] = 1
        _plot_detrend(X, y, w)

    return y


def _plot_detrend(x, y, w):
    """Plot detrending results."""
    import matplotlib.pyplot as plt
    from matplotlib.gridspec import GridSpec

    n_times = x.shape[0]
    n_chans = x.shape[1]

    f = plt.figure()
    gs = GridSpec(4, 1, figure=f)
    ax1 = f.add_subplot(gs[:3, 0])
    lines = ax1.plot(x, label="original", color="C0")
    plt.setp(lines[1:], label="_")
    lines = ax1.plot(y, label="detrended", color="C1")
    plt.setp(lines[1:], label="_")
    ax1.set_xlim(0, n_times)
    ax1.set_xticklabels("")
    ax1.set_title("Robust detrending")
    ax1.legend(fontsize="smaller")

    ax2 = f.add_subplot(gs[3, 0])
    ax2.pcolormesh(w.T, cmap="Greys")
    ax2.set_yticks(np.arange(0, n_chans) + 0.5)
    ax2.set_yticklabels([f"ch{i}" for i in np.arange(n_chans)])
    ax2.set_xlim(0, n_times)
    ax2.set_ylabel("ch. weights")
    ax2.set_xlabel("samples")
    plt.show()


def create_masked_weight(x, events, tmin, tmax, sfreq):
    """Output a weight matrix for trial-masked detrending.

    Creates a  (n_times, n_channels) weight matrix with masked
    periods (value of zero) in order to mask the trials of interest during
    detrending [1]_.

    Parameters
    ----------
    x : ndarray, shape=(n_times, n_channels)
        Raw data matrix.
    events : ndarray, shape=(n_events)
        Time samples of the events.
    tmin : float
        Start time before event (in seconds).
    tmax : float
        End time after event (in seconds).
    sfreq : float
        The sampling frequency of the data.

    Returns
    -------
    weights : ndarray, shape=(n_times, n_channels)
        Weight for each channel and each time sample (zero is masked).

    References
    ----------
    .. [1] van Driel, J., Olivers, C. N., & Fahrenfort, J. J. (2021). High-pass
       filtering artifacts in multivariate classification of neural time series
       data. Journal of Neuroscience Methods, 352, 109080.

    """
    if x.ndim != 2:
        raise ValueError("The shape of x must be (n_times, n_channels)")

    weights = np.ones(x.shape)
    for e in events:
        weights[int(e + tmin * sfreq): int(e + tmax * sfreq + 1), :] = 0

    return weights
